"""RGSW encryption scheme."""

import dataclasses
import functools

import jax
import jax.numpy as jnp
from jaxite.jaxite_lib import decomposition
from jaxite.jaxite_lib import encoding
from jaxite.jaxite_lib import matrix_utils
from jaxite.jaxite_lib import parameters
from jaxite.jaxite_lib import random_source
from jaxite.jaxite_lib import rlwe


@dataclasses.dataclass
class RgswPlaintext:
  """An RGSW plaintext is an unsigned integer and a polynomial modulus degree.

  Note: in the TFHE paper, RGSW plaintext is a polynomial in Z[X] / (X^N + 1),
  where N = `modulus_degree` below. However, in TFHE the only use of RGSW is to
  encrypt the individual bits of an LWE secret key, a binary vector. In this
  case, the RgswPlaintext is in {0, 1} and we can safely restrict the plaintext
  type to an unsigned integer.
  """

  # The degree N of the ring modulus polynomial.
  modulus_degree: int

  # The unsigned integer representing the plaintext.
  message: jnp.uint32

  def __str__(self) -> str:
    return str(self.message)


@dataclasses.dataclass
class RgswCiphertext:
  """An RGSW ciphertext."""

  # A 3-dimensional array, where each row contains an RLWE encryption
  # of a particular bit of the message. Recall, an RLWE encryption is a list of
  # polynomials, so each (i,j) entry corresponds to a polynomial, with the third
  # dimension being the coefficients of the polynomial in order of increasing
  # degree.
  message: jnp.ndarray

  # the log of the modulus q of the polynomial coefficients
  log_coefficient_modulus: int

  # the degree N of the ring modulus polynomial.
  modulus_degree: int

  @property
  def coefficient_modulus(self) -> jnp.uint32:
    return jnp.uint32(2) ** self.log_coefficient_modulus

  def __len__(self):
    return self.message.shape[0]

  def pretty_print_polynomial(self, row: jnp.ndarray) -> str:
    # this does not need to be fast because it will only be used in development.
    s = ' + '.join(
        f'{coeff} x^{power}' for (power, coeff) in enumerate(row) if coeff != 0
    )
    return s if s else '0'

  def __str__(self) -> str:
    s = [
        [self.pretty_print_polynomial(e) + ',' for e in row]
        for row in self.message
    ]
    lens = [max(map(len, col)) for col in zip(*s)]
    fmt = '\t'.join('{{:{}}}'.format(x) for x in lens)
    table = [fmt.format(*row) for row in s]
    return '\n[\n  ' + '\n  '.join(table) + '\n]\n'

  def __repr__(self) -> str:
    return str(self)


@dataclasses.dataclass
class RgswSecretKey:
  """This is a wrapper around RlweSecretKey."""

  # the underlying RLWE secret key
  key: rlwe.RlweSecretKey

  def to_rlwe_secret_key(self) -> rlwe.RlweSecretKey:
    return self.key

  def data_at_index(self, i: int) -> jnp.uint32:
    return self.key.data[i]


def gen_key(
    params: parameters.SchemeParameters, prg: random_source.RandomSource
) -> RgswSecretKey:
  """Generate an RGSW secret key."""
  return RgswSecretKey(key=rlwe.gen_key(params, prg))


def key_from_rlwe(rlwe_key: rlwe.RlweSecretKey) -> RgswSecretKey:
  """Convert an RLWE secret key to the corresponding RGSW secret key."""
  return RgswSecretKey(key=rlwe_key)


def encrypt(
    plaintext: RgswPlaintext,
    sk: RgswSecretKey,
    decomposition_params: decomposition.DecompositionParameters,
    prg: random_source.RandomSource,
) -> RgswCiphertext:
  """Create an RGSW ciphertext."""
  k = num_blocks = sk.key.rlwe_dimension
  levels = decomposition_params.level_count
  rlwe_sk = sk.to_rlwe_secret_key()

  # Because of the way the vmaps are set up to map individual rows to each call
  # to encrypt_block and encrypt_and_modify_one_row, we organize the randomness
  # into a matching shape, so that the output of ai_samples[block, level] is in
  # the right shape to be used for a single RLWE encryption. Nb., it needs only
  # k random samples, not k+1 like the output array's shape, because the extra 1
  # comes from the b term of the RLWE encryption.
  ai_samples = prg.uniform(
      shape=(num_blocks + 1, levels, k, rlwe_sk.modulus_degree),
      dtype=jnp.uint32,
  )
  error_samples = prg.rounded_normal(
      shape=(num_blocks + 1, levels),
      dtype=jnp.uint32,
  )

  ciphertext = jit_encrypt(
      plaintext.message,
      rlwe_sk.data,
      ai_samples,
      error_samples,
      decomposition_params,
      num_blocks,
      rlwe_sk.log_coefficient_modulus,
      rlwe_sk.modulus_degree,
  )

  return RgswCiphertext(
      message=ciphertext,
      log_coefficient_modulus=rlwe_sk.log_coefficient_modulus,
      modulus_degree=plaintext.modulus_degree,
  )


@functools.partial(
    jax.jit,
    static_argnames=(
        'decomposition_params',
        'num_blocks',
        'log_coefficient_modulus',
        'modulus_degree',
    ),
)
def jit_encrypt(
    plaintext: jnp.ndarray,
    rlwe_sk: jnp.ndarray,
    ai_samples: jnp.ndarray,
    error_samples: jnp.ndarray,
    decomposition_params: decomposition.DecompositionParameters,
    num_blocks: int,
    log_coefficient_modulus: int,
    modulus_degree: int,
) -> jnp.ndarray:
  """Create an RGSW ciphertext."""
  levels = decomposition_params.level_count
  log_base = decomposition_params.log_base
  levels_range = jnp.arange(1, levels + 1, dtype=jnp.uint32)
  block_range = jnp.arange(num_blocks + 1, dtype=jnp.uint32)
  zero_to_encrypt = jnp.array([0], dtype=jnp.uint32)

  # In the Joye paper, p. 19, we're computing Z = m * G^T,
  # where G^T is defined on p. 17 as the block diagonal matrix
  #
  # ( 1/B                        )
  # ( 1/B^2                      )
  # ( ...                        )
  # ( 1/B^l                      )
  # (        1/B                 )
  # (        1/B^2               )
  # (        ...                 )
  # (        1/B^l               )
  # (               .            )
  # (                 .          )
  # (                   .        )
  # (                      1/B   )
  # (                      1/B^2 )
  # (                      ...   )
  # (                      1/B^l )
  #
  # However, in our case we're working in 32-bit uints, so each value is
  # multiplied by B^l == 2**(log_base*levels), which makes each block above
  # into (B^{l-1}, B^{l-2}, ..., 1).
  #
  # The elements of the matrix are polynomials, which here are represented as as
  # vectors of coefficients.  This requires us to make a 3-dimensional array.
  # The t-th row contains an RLWE encryption, which is a list of polynomials,
  # and the (t,s) entry contains the coefficients of a single polynomial.
  #
  # The jax magic below splits the generation of the matrix above into blocks
  # processed in the same order as in the diagram above.
  #
  #  - encrypt_and_modify_one_row computes the output for a single row of the
  #    matrix above, i.e., one RLWE encryption of zero + the decomposed
  #    plaintext.
  #  - encrypt_block computes vmaps each row computation for a block.
  #  - the final vmap maps across blocks.

  @jax.jit
  def encrypt_and_modify_one_row(
      ai_samples,
      error_sample,
      level,
      block,
      plaintext_message,
  ):
    # For each row of the matrix m * G^T above, we generate a fresh RLWE
    # encryption of zero, and because the message in our case is always an
    # integer, we only need to modify the constant term of the appropriate
    # polynomial, adding this decomposed_pt to it.
    decomposed_pt = plaintext_message << (
        log_coefficient_modulus - log_base * level
    )
    rlwe_zero = rlwe.jit_encrypt(
        zero_to_encrypt,
        rlwe_sk,
        ai_samples,
        error_sample,
        log_coefficient_modulus,
    )

    # Modify the constant coefficient by adding the decomposed plaintext.
    modified_poly = rlwe_zero[block]
    modfied_first_coeff = modified_poly[0] + decomposed_pt
    modified_poly = modified_poly.at[0].set(modfied_first_coeff)
    return rlwe_zero.at[block].set(modified_poly)

  @jax.jit
  def encrypt_block(ai_samples, error_samples, block, plaintext_message):
    return jax.vmap(
        encrypt_and_modify_one_row,
        in_axes=(0, 0, 0, None, None),
        out_axes=0,
    )(ai_samples, error_samples, levels_range, block, plaintext_message)

  ciphertext = jax.vmap(
      encrypt_block,
      in_axes=(0, 0, 0, None),
      out_axes=0,
  )(ai_samples, error_samples, block_range, plaintext)

  return ciphertext.reshape(
      (levels * (num_blocks + 1), (num_blocks + 1), modulus_degree)
  )


def decrypt(
    ciphertext: RgswCiphertext,
    decomposition_params: decomposition.DecompositionParameters,
    sk: RgswSecretKey,
) -> RgswPlaintext:
  """Decrypts an RGSW ciphertext.

  The approach is to isolate the modified polynomial in a particular
  RgswCiphertext row, and use it to reconstruct the message m.

  - pick `row` = (a_1, ... a_j + mB^{s-1}, ..., a_k, b), * i.e., with entry `j`
      being the modified entry from the encryption step, and the row is chosen
      such that s = levels is maximal.
  - Let b  = sum_{i=1}^k s_i * a_i + e (since it was an RLWE encryption of 0)
  - Let b' = sum_{i=1}^k s_i * row_i = b - e + s_j m B^{s-1}
  - Note s_j must be 1 to recover m, which may affect the choice of row.
  - decomposed_pt_with_noise = b' - b = s_j m B^{s-1} - e = m B^{s-1} - e
                             = decomposed_pt - e
  - pt_with_noise = round(decomposed_pt_with_noise)
  - pt_guess = pt_with_noise[0] mod coeff_modulus
  - pt_guess ~~ plaintext

  Args:
    ciphertext: RGSW ciphertext to decrypt
    decomposition_params: parameters for decomposing the plaintext
    sk: RGSW secret key to use to decrypt

  Returns:
    RgswPlaintext: decrypted plaintext
  """
  k = sk.key.rlwe_dimension
  rlwe_sk = sk.to_rlwe_secret_key()
  log_base = decomposition_params.log_base
  omega = rlwe_sk.log_coefficient_modulus

  # this will fail if there is no such entry, which is necessary or else the
  # decryption will fail.
  try:
    sk_index = next(index for index in range(k) if rlwe_sk.data[index, 0] == 1)
  except StopIteration:
    raise ValueError(
        'Expected SK to have an entry with a nonzero '
        f'constant term, got: {rlwe_sk}'
    ) from None

  row_index = decomposition_params.level_count * sk_index
  ciphertext_row = ciphertext.message[row_index]

  # b = sum{i=0}^k s_i * a_i + e (since it was an RLWE encryption of 0)
  b = ciphertext_row[k]
  # b' = sum_{i=1}^k s_i * row_i
  inner_product = jnp.sum(
      jnp.array(
          [
              matrix_utils.poly_mul(a, s)
              for a, s, in zip(ciphertext_row[:k], rlwe_sk.data)
          ],
          dtype=jnp.uint32,
      ),
      axis=0,
  )
  # decomposed_pt_with_noise = b' - b
  #                          = decomposed_pt - e
  decomposed_pt_with_noise = inner_product - b

  # Round and shift back from the PowersOfB encoding of the original input.
  # the ciphertext row was chosen so that we're always working with the
  # largest block of the decomposition, i.e., corresponding to level=1
  # so this would be omega - log_base * level in the encryption step.
  shift_length = omega - log_base
  pt_with_noise = encoding.round_to_power_of_2(
      decomposed_pt_with_noise, shift_length
  )
  pt_with_noise = pt_with_noise >> shift_length
  # pt_guess = pt_with_noise[0] mod coeff_modulus
  pt_guess = jnp.mod(pt_with_noise[0], 2**log_base)

  return RgswPlaintext(message=pt_guess, modulus_degree=rlwe_sk.modulus_degree)



"""Helpers for type conversions in jaxite_bool client code."""

from typing import List, Optional


def bit_slice_to_u8(bit_slice: List[bool]) -> int:
  """Given a bit slice of length 8, returns a base-10 int representation."""
  if len(bit_slice) != 8:
    raise ValueError(f'Expected an 8-bit representation but got: {bit_slice}.')
  result = 0
  for i in range(8):
    result |= (int(bit_slice[i])) << i
  return result


def u8_to_bit_slice(input_int: int) -> List[bool]:
  """Given an integer [0, 255], returns a bitwise representation."""
  if input_int < 0 or input_int > 255:
    raise ValueError(f'Expected a u8, but got: {input_int}.')
  result: List[bool] = [False] * 8
  for i in range(8):
    result[i] = ((input_int >> i) & 1) != 0
  return result


def u8_list_to_bit_slice(input_list: List[int]) -> List[bool]:
  """Given a list of u8 values, returns a flattened bitwise representation."""
  if not input_list:
    raise ValueError('Expected a non-empty list of u8 values.')
  result: List[bool] = []
  for i in input_list:
    result.extend(u8_to_bit_slice(i))
  return result


def bit_slice_to_bytes(bit_slice: List[bool]) -> bytes:
  """Given a bitwise representation, returns an ASCII bytes object."""
  if not bit_slice:
    raise ValueError('Expected a non-empty bit slice.')
  result: List[int] = []

  char_bit_slice: List[bool] = [False] * 8
  for i, entry in enumerate(bit_slice):
    if (i + 1) % 8 == 0:
      result.append(bit_slice_to_u8(char_bit_slice))
    char_bit_slice[i % 8] = entry

  return bytes(result)


def str_to_cleartext(
    text: str,
    static_len: Optional[int] = None,
    padding_byte: bytes = bytes(' ', 'ascii'),
) -> List[bool]:
  cleartext_bytes = list(bytes(text, 'ascii'))
  if static_len:
    cleartext_bytes.extend(list(padding_byte) * (static_len - len(text)))
  return u8_list_to_bit_slice(cleartext_bytes)


def cleartext_to_str(cleartext: List[bool]) -> str:
  return bit_slice_to_bytes(cleartext).decode('ascii')



"""Security parameters for binary encoding for CGGI schemes."""
import functools

from jaxite.jaxite_bool import bool_encoding
from jaxite.jaxite_bool import lut as bool_lut
from jaxite.jaxite_lib import decomposition
from jaxite.jaxite_lib import encoding
from jaxite.jaxite_lib import lwe
from jaxite.jaxite_lib import parameters
from jaxite.jaxite_lib import random_source
from jaxite.jaxite_lib import rlwe
from jaxite.jaxite_lib import types

ENCODING_PARAMS = bool_encoding.ENCODING_PARAMS


class Parameters:
  """TFHE Boolean gate shared parameters."""

  def __init__(
      self,
      scheme_params: parameters.SchemeParameters,
      ks_decomp_params: decomposition.DecompositionParameters,
      bs_decomp_params: decomposition.DecompositionParameters,
  ) -> None:
    self._scheme_params = scheme_params
    self._ks_decomp_params = ks_decomp_params
    self._bs_decomp_params = bs_decomp_params
    self._lut_cache = bool_lut.LutCache(self._scheme_params)

  @property
  def scheme_params(self) -> parameters.SchemeParameters:
    return self._scheme_params

  @property
  def ks_decomp_params(self) -> decomposition.DecompositionParameters:
    return self._ks_decomp_params

  @property
  def bs_decomp_params(self) -> decomposition.DecompositionParameters:
    return self._bs_decomp_params

  def _noiseless_embedding(
      self, cleartext: types.LweCleartext
  ) -> types.LweCiphertext:
    return lwe.noiseless_embedding(
        encoding.encode(cleartext, ENCODING_PARAMS),
        self.scheme_params.lwe_dimension,
    )

  @functools.cached_property
  def noiseless_true(self) -> types.LweCiphertext:
    return self._noiseless_embedding(bool_encoding.CLEARTEXT_TRUE)

  @functools.cached_property
  def noiseless_false(self) -> types.LweCiphertext:
    return self._noiseless_embedding(bool_encoding.CLEARTEXT_FALSE)

  def lut_poly(self, num_inputs: int, truth_table: int) -> rlwe.RlweCiphertext:
    return self._lut_cache.lut_poly(
        num_inputs=num_inputs, lut_as_int=truth_table
    )

  def lut_poly_by_name(self, name: str) -> rlwe.RlweCiphertext:
    return self._lut_cache.lut_poly_by_name(name)

  def lut(self, num_inputs: int, truth_table: int) -> bool_lut.LookUpTable:
    return self._lut_cache.lut(num_inputs=num_inputs, lut_as_int=truth_table)

  def lut_by_name(self, name: str) -> bool_lut.LookUpTable:
    return self._lut_cache.lut_by_name(name)


def get_params_for_128_bit_security() -> Parameters:
  """Returns boolean scheme params for 128 bit security."""
  return Parameters(
      SCHEME_PARAMS_128_BIT_SECURITY,
      KSK_DECOMP_PARAMS_128_BIT_SECURITY,
      BSK_DECOMP_PARAMS_128_BIT_SECURITY,
  )


def get_lwe_rng_for_128_bit_security(
    seed: int,
) -> random_source.PseudorandomSource:
  """Returns lwe rng for 128 bit security."""
  # This parameter was selected to satisfy 128-bit security while giving
  # efficient performance. See go/fhe-params for more details.
  return random_source.PseudorandomSource(seed=seed, normal_std=2**14)


def get_rlwe_rng_for_128_bit_security(
    seed: int,
) -> random_source.PseudorandomSource:
  """Returns rlwe rng for 128 bit security."""
  # This parameter was selected to satisfy 128-bit security while giving
  # efficient performance. See go/fhe-params for more details.
  return random_source.PseudorandomSource(seed=seed, normal_std=256)


# Scheme params and decomposition params for 128 bit security.
SCHEME_PARAMS_128_BIT_SECURITY = parameters.SchemeParameters(
    lwe_dimension=800,
    plaintext_modulus=4294967296,
    rlwe_dimension=2,
    polynomial_modulus_degree=512,
)
# Note: The following condition needs to be met for decomposition parameters
# in order to avoid loss of precision:
# (2^decomposition_log_base)^decomposition_level_count == plaintext_modulus.
BSK_DECOMP_PARAMS_128_BIT_SECURITY = decomposition.DecompositionParameters(
    log_base=4, level_count=6
)
# These parameters give an approximate decomposition. See go/keyswitch-error.
KSK_DECOMP_PARAMS_128_BIT_SECURITY = decomposition.DecompositionParameters(
    log_base=4, level_count=5
)


def get_params_for_test() -> Parameters:
  """Returns boolean scheme params for tests."""
  return Parameters(
      TEST_SCHEME_PARAMS, TEST_KSK_DECOMP_PARAMS, TEST_BSK_DECOMP_PARAMS
  )


def get_rng_for_test(seed: int) -> random_source.PseudorandomSource:
  """Returns rng for tests."""
  return random_source.PseudorandomSource(
      uniform_bounds=(0, 2**32), normal_std=1, seed=seed
  )


# Scheme params and decomposition params used in unit tests.
TEST_SCHEME_PARAMS = parameters.SchemeParameters(
    lwe_dimension=4,
    plaintext_modulus=2**32,
    rlwe_dimension=1,
    polynomial_modulus_degree=512,
)
# Note: The following condition needs to be met for decomposition parameters:
# (2^decomposition_log_base)^decomposition_level_count == plaintext_modulus.
TEST_BSK_DECOMP_PARAMS = decomposition.DecompositionParameters(
    log_base=4, level_count=8
)
TEST_KSK_DECOMP_PARAMS = TEST_BSK_DECOMP_PARAMS



"""Logic for encoding and decoding a cleartext for use in TFHE."""

import dataclasses
from typing import Union

import jax.numpy as jnp
from jaxite.jaxite_lib import types


@dataclasses.dataclass
class EncodingParameters:
  """The parameters for encoding a cleartext for use in TFHE.

  See g.p.p.fhe.jaxite.jaxite_lib.encoding.encode for the structure of the
  encoding.
  """

  # total number of bits in a plaintext
  total_bit_length: int

  # the number of bits to use to store the message
  message_bit_length: int

  # The number of bits to use for padding, stored in the most sigificant bits
  # of a plaintext. This is needed to avoid the negation that occurs when
  # blind_rotate ends up indexing past the degree of the test polynomial.
  # See go/cggi#padding-and-blind-rotate for more details.
  # TODO(b/238643320): determine if we can/should automatically set this based
  # on the scheme parameters.
  padding_bit_length: int

  # the number of bits used for error; stored in the LSBs
  error_bit_length: int = dataclasses.field(init=False)

  # minimum allowed message value
  message_min: types.LweCleartext = dataclasses.field(init=False)

  # maximum allowed message value
  message_max: types.LweCleartext = dataclasses.field(init=False)

  # range of allowed message values
  message_range: types.LweCleartext = dataclasses.field(init=False)

  def __post_init__(self) -> None:
    # verify bit lengths are of appropriate size
    if self.total_bit_length < 1:
      raise ValueError('Total bit length must be >= 1.')
    if self.message_bit_length < 1:
      raise ValueError('Message bit length must be >= 1.')
    if self.padding_bit_length < 0:
      raise ValueError('Padding bit length must be >= 0.')

    # compute error bit length based on available space
    occupied_bits = self.message_bit_length + self.padding_bit_length
    remaining_bits = self.total_bit_length - occupied_bits

    if remaining_bits < 0:
      raise ValueError(f'Total bit length of {self.total_bit_length} exceeded.')

    self.error_bit_length = remaining_bits
    self.message_min = 0
    self.message_max = (1 << self.message_bit_length) - 1
    self.message_range = self.message_max - self.message_min


def encode(
    message: Union[types.LweCleartext, jnp.ndarray], params: EncodingParameters
) -> Union[types.LwePlaintext, jnp.ndarray]:
  """Encode a plaintext or array of plaintexts for use in a TFHE ciphertext.

  The bits are organized so that the top bits are padding for overflow,
  followed by message bits, followed by space left for noise. E.g.,

           00 0101101 00000
      padding message noise

  This method also works on arrays, for use in encoding the coefficients
  of a test polynomial.

  Args:
    message: the cleartext message.
    params: the parameters of the encoding.

  Returns:
    The encoded message or list of messages.

  Raises:
    ValueError: In the event that `message` is outside of the message bounds as
    specified in `params`.
  """
  dtype = types.LwePlaintext
  msg_arr = jnp.atleast_1d(message).astype(dtype)
  if (msg_arr > dtype(params.message_max)).any() or (
      msg_arr < dtype(params.message_min)
  ).any():
    raise ValueError(
        f'{message} is outside of allowable bounds '
        f'[{params.message_min}, {params.message_max}].'
    )

  return message << dtype(params.error_bit_length)


def decode(
    plaintext: types.LwePlaintext, params: EncodingParameters
) -> types.LweCleartext:
  """Decode a plaintext.

  Args:
    plaintext: the encoded plaintext message.
    params: the parameters of the encoding.

  Returns:
    The cleartext message.
  """
  shifted = remove_noise(plaintext, params) >> params.error_bit_length

  # In cases where the params.total_bit_length < 32, operations like modulus
  # switching can result in a plaintext that spills into the range of bits that
  # are outside the highest bit in the encoding. E.g., if the cleartext=0 and
  # message_bits=3, modulus switching can result in encrypting the "cleartext"
  # 8, which is only correct if you ignore the top bit. To deal with cases like
  # this (and really, this matters only for tests where we are not using all the
  # bits of the message space), we compute a final modulus with respect to the
  # message space size.
  message_space_mask = (1 << params.message_bit_length) - 1
  return types.LweCleartext(shifted & message_space_mask)


def remove_noise(
    inp: jnp.ndarray, encoding_params: EncodingParameters
) -> jnp.ndarray:
  """Uses encoding_params to round the inp to remove noise."""
  return round_to_power_of_2(inp, encoding_params.error_bit_length)


def round_to_power_of_2(arr: jnp.ndarray, log_pow_of_2: int) -> jnp.ndarray:
  """Rounds to the nearest multiple of a given power of 2."""
  # This bit determines whether to round up or down
  round_up_or_down_bit = log_pow_of_2 - 1
  lowest_unrounded_bit = log_pow_of_2

  # Shift down to clear all the bits that are rounded off, optionally add 1 to
  # round up, then shift back up.
  round_up = jnp.bitwise_and(jnp.right_shift(arr, round_up_or_down_bit), 1)
  return jnp.left_shift(
      jnp.right_shift(arr, lowest_unrounded_bit) + round_up,
      lowest_unrounded_bit,
  )


def extract_noise(
    plaintext: types.LwePlaintext, encoding_params: EncodingParameters
) -> int:
  """Extracts the noise bits of a plaintext as a (signed) int."""
  rounded = remove_noise(plaintext, encoding_params)
  return int(jnp.int32(plaintext)) - int(jnp.int32(rounded))



"""Util file for operations over matrices."""

import functools

import jax
import jax.numpy as jnp


@jax.jit
def integer_div(values: jnp.ndarray, divisor: jnp.uint32) -> jnp.ndarray:
  """Performs integer division with rounding for positive integers X, d.

  Args:
    values: A vector of values to be divided by `divisor`.
    divisor: The divisor.

  Returns:
    An `jnp.ndarray` vectorizing the division across `values`.
  """
  # If not for overflow, we could compute as: (X + d//2)//d. Adding d//2 to the
  # numerator prior to dividing rounds the result up iff the fractional part of
  # X/d is >= 0.5. This avoids floating point error due to conversion to float32
  # during division.

  # However, the above fails with uint32 for values of X, d close to 2**32. So
  # we split the division in two parts. First get the floor of the quotient,
  # X//d.
  quotient_floor = values // divisor

  # Then compute the quotient of the remainder (X%d) with rounding, which is
  # guaranteed to have a numerator small enough to avoid overflow, and always
  # results in either 0 or 1.
  rounded_fractional_part = ((values % divisor) + divisor // 2) // divisor

  # Adding them together effectively rounds up or down.
  return quotient_floor + rounded_fractional_part


@functools.partial(jax.jit, static_argnames="poly_mod_deg")
def x_power_n_minus_1(n: jnp.uint32, poly_mod_deg: jnp.uint32) -> jnp.ndarray:
  """Construct a polynomial of the form x^d - 1 for an input power d.

  The polynomial is reduced modulo (x^poly_mod_deg + 1). If n == 0, the zero
  polynomial is returned.

  Args:
    n: the nonzero exponent
    poly_mod_deg: the degree of the polynomial modulus to use for reduction. The
      output polynomial is represented by its coefficients as an array of length
      poly_mod_deg+1.

  Returns:
    An array representing the polynomial.
  """
  degree = n % (2 * poly_mod_deg)
  flip = degree // poly_mod_deg
  reduced_degree = degree % poly_mod_deg
  zeros = jnp.zeros(poly_mod_deg, dtype=jnp.uint32)
  return zeros.at[reduced_degree].set((-1) ** flip) - zeros.at[0].set(1)


@jax.jit
def int32_to_int8_arr(arr: jnp.array) -> jnp.array:
  """Decompose an int32 matrix into u8s."""
  return jax.lax.bitcast_convert_type(arr, new_dtype=jnp.uint8)


@jax.jit
def i32_as_u8_matmul(lhs: jnp.array, rhs: jnp.array) -> jnp.array:
  """Multiply an (n,) by an (n, k) i32 matrix using only i8 ops."""
  if lhs.ndim != 1 or rhs.ndim != 2:
    raise ValueError(
        "lhs be 1-dim and rhs must be 2-dim, "
        f"but they were: lhs={lhs.shape}, rhs={rhs.shape}"
    )

  lhs = int32_to_int8_arr(lhs)
  rhs = int32_to_int8_arr(rhs)

  i8_products = jnp.einsum(
      "np,nkq->kpq",
      lhs,
      rhs,
      preferred_element_type=jnp.int32,
  )
  shift_factors = jnp.array(
      [
          [0, 8, 16, 24],
          [8, 16, 24, 32],
          [16, 24, 32, 40],
          [24, 32, 40, 48],
      ],
      dtype=jnp.int32,
  )
  return jnp.sum(i8_products << shift_factors, axis=(1, 2))


# https://jax.readthedocs.io/en/latest/_autosummary/jax.numpy.tril.html
# For n=3, generates the following
# [[ 1 -1 -1]
#  [ 1  1 -1]
#  [ 1  1  1]]
@jax.named_call
def _generate_sign_matrix(n: int) -> jnp.ndarray:
  """Generates a sign matrix with 1s below the diagonal and -1 above."""
  up_tri = jnp.tril(jnp.ones((n, n), dtype=int), 0)
  low_tri = jnp.triu(jnp.ones((n, n), dtype=int), 1) * -1
  return up_tri + low_tri


@jax.named_call
@jax.jit
def toeplitz(x: jnp.ndarray) -> jnp.ndarray:
  """Generates a cyclic matrix with each row of the input shifted.

  For input: [1, 9, 2], generates the following matrix:
  [[1 9 2]
   [2 1 9]
   [9 2 1]]

  Args:
    x: the 1D array to shift of length n

  Returns:
    A 2D matrix of shape (n, n), with row i containing the input rolled
    rightward i times.
  """
  if len(x.shape) == 1:
    return toeplitz(x[:, None])
  n = x.shape[-2]
  m = x.shape[-1]
  if m == n:
    return x.transpose()
  if m > n:
    return x[..., :n].transpose()
  r = jnp.roll(x, m, axis=-2)
  return toeplitz(jnp.concatenate([x, r], axis=-1))


@jax.named_call
@jax.jit
def toeplitz_poly_mul(a: jnp.array, b: jnp.array) -> jnp.ndarray:
  """Computes a poly multiplication mod (X^N + 1) where N = len(a)."""
  multiplier = _generate_sign_matrix(len(a))
  left_matrix = multiplier * toeplitz(a).transpose()
  return jnp.matmul(b, left_matrix.transpose())


@jax.named_call
@jax.jit
def poly_mul(a: jnp.array, b: jnp.array) -> jnp.ndarray:
  """Computes a poly multiplication mod (X^N + 1) where N = len(a)."""
  return toeplitz_poly_mul(a, b)


@jax.named_call
@functools.partial(jax.jit, static_argnames="log_modulus")
def monomial_mul(poly: jnp.array, degree: int, log_modulus: int) -> jnp.ndarray:
  """Computes `poly * X^degree mod (X^N + 1)` where N = len(poly).

  Args:
    poly: a polynomial to be multiplied by X^degree
    degree: the degree of the monomial, X^degree, which may be positive or
      negative. A negative degree is interpreted as a monomial division.
    log_modulus: the log of the coefficient modulus of the polynomial

  Returns:
    The result of the polynomial `poly` multiplied by monomial of degree
    `degree`, mod `2**log_modulus` and `X^N + 1`.

  Example:
     Args: poly = [1, 2, 3], degree = 2, log_modulus = 8
     Returns: [254, 253, 1]

  This takes advantage of the negacyclic property of polynomials mod (X^N + 1),
  as described in Example 9 (p.15) of the Joye TFHE Explainer
  (https://eprint.iacr.org/2021/1402.pdf). At each mulitplication by X, the
  polynomial coefficients are circularly shifted one position to the right and
  the entering coefficient is negated.

  Operations:
    1. After a multiplication by X^{2N}, the polynomial is unchanged. Therefore,
    we can take degree % 2N.
    2. After a multiplication by X^N, all the coefficients of the polynomial
    have been negated but the positions are unchanged.
    Therefore, if degree % 2N > N, negate the coefficients and take degree % N.
    3. Divide the coefficients into "wrapped" (negated) and "unwrapped"
    (unmodified) coefficients, and rearrange the indices accordingly:
      - unwrapped coefficients are moved from the start to the end of the
        polynomial
      - wrapped coefficients are negated and moved to the start of the
        polynomial
  """
  n = poly.shape[0]

  # After a multiplication by X^{2N}, the polynomial is unchanged.
  degree = degree % (2 * n)
  flip = degree // n
  shift = degree % n

  # equivalent to "if degree >= n: poly = -poly" because 0 <= degree < 2N
  poly = jnp.uint32((-1) ** flip) * poly
  rolled = jnp.roll(poly, shift)

  # trick: generate an array like [1, 1, ..., 1, -1, -1, ..., -1]
  # and rolling that gives the right argument to use in an element-wise product
  # with the tail truncated.
  ones = jnp.ones(n, dtype=jnp.uint32)
  sign = jnp.roll(jnp.concatenate([ones, -ones]), shift)
  output = rolled * sign[:n]

  if 0 < log_modulus < 32:
    output = jnp.mod(output, jnp.uint32(2) ** log_modulus)

  return output


monomial_mul_list = jax.vmap(monomial_mul, in_axes=(0, None, None), out_axes=0)

# in_axes = (0, 0) means that we apply poly_mul to each row of two input
# matrices. out_axes = (0, ) means the output arrays are stacked one per row
poly_mul_list = jax.vmap(poly_mul, in_axes=(0, 0), out_axes=0)

# poly_mul_const_matrix(poly, list) applies poly_mul(poly, x) to each value x
# of `list`.
poly_mul_const_list = jax.vmap(poly_mul, in_axes=(None, 0), out_axes=0)

# poly_mul_const_matrix(poly, matrix) applies poly_mul(poly, x) to each value x
# of `matrix`.
poly_mul_const_matrix = jax.vmap(
    poly_mul_const_list, in_axes=(None, 0), out_axes=0
)


def poly_dot_product(
    poly_vec1: jnp.ndarray, poly_vec2: jnp.ndarray
) -> jnp.ndarray:
  """Compute a dot product of two vectors of polynomials."""
  return jnp.sum(poly_mul_list(poly_vec1, poly_vec2), axis=0).astype(jnp.uint32)



"""Test utilities."""

import itertools
import math
from typing import Optional

import jax.numpy as jnp
from jaxite.jaxite_bool import bool_params
from jaxite.jaxite_lib import encoding
from jaxite.jaxite_lib import lwe
from jaxite.jaxite_lib import parameters
from jaxite.jaxite_lib import rlwe

DEFAULT_ENCODING_PARAMS = encoding.EncodingParameters(
    total_bit_length=26, message_bit_length=14, padding_bit_length=2
)
ENCODING_PARAMS_128_BIT_SECURITY = encoding.EncodingParameters(
    total_bit_length=32, message_bit_length=6, padding_bit_length=1
)

LWE_RNG_128_BIT_SECURITY = bool_params.get_lwe_rng_for_128_bit_security(2)
RLWE_RNG_128_BIT_SECURITY = bool_params.get_rlwe_rng_for_128_bit_security(2)
SCHEME_PARAMS_128_BIT_SECURITY = bool_params.SCHEME_PARAMS_128_BIT_SECURITY

BSK_DECOMP_PARAMS_128_BIT_SECURITY = (
    bool_params.BSK_DECOMP_PARAMS_128_BIT_SECURITY
)
KSK_DECOMP_PARAMS_128_BIT_SECURITY = (
    bool_params.KSK_DECOMP_PARAMS_128_BIT_SECURITY
)


class ConsistencyChecker:
  """A helper for checking consistency of intermediate values across runs."""

  def __init__(self):
    self.insertion_order = []
    self.records = dict()

  def store(self, name: str, value: jnp.ndarray) -> None:
    value = value.tolist()
    if name in self.records:
      self.records[name].append(value)
    else:
      self.insertion_order.append(name)
      self.records[name] = [value]

  def check(self, asserter: ... = None) -> None:
    assert asserter is not None
    for name in self.insertion_order:
      print(f"Consistency check for {name}")
      values = self.records[name]
      for x, y in itertools.combinations(values, 2):
        asserter(x, y)


def assert_safe_modulus_switch(
    mod_degree: int, message_bits: int, lwe_dimension: int
):
  """Guard against corruption of the message due to modulus switch error.

  The ratio of mod_degree to 2**message_bits defines the size of a block in the
  test polynomial. Half that is the radius of the block, and that defines the
  error allowed in the post-modulus-switched ciphertext before it corrupts the
  message. Modulus switch can add up to sqrt(lwe_dim * log(log(lwe_dim))) error,
  and this implies a limit on how large lwe_dim can be before mod_degree must
  increase.

  Args:
    mod_degree: the degree of the RLWE polynomial modulus
    message_bits: the log2 of the size of the message space.
    lwe_dimension: the number of samples used in the LWE ciphertexts.

  Returns:
    None. Assertion failure if params are set improperly.
  """
  test_poly_block_radius = 0.5 * mod_degree / 2**message_bits
  # add one because the rounding of the `b` term also introduces error while
  # lwe_dimension tracks the number of samples only.
  n = 1 + lwe_dimension
  # TODO(b/242187167): show that the bounds can be tightened to sqrt(n).
  mod_switch_extra_error = math.sqrt(n)
  assert mod_switch_extra_error <= test_poly_block_radius, (
      f"Modulus switch error may be as large as {mod_switch_extra_error}, "
      "but the test polynomial repetition blocks "
      f"only have radius {test_poly_block_radius}. "
      "This risks the error from modulus switch corrupting the message. "
      f"(mod_degree={mod_degree}, lwe_dim={lwe_dimension}, "
      f"message_space={2**message_bits})"
  )


class MidBootstrapDecrypter:
  """A helper for decrypting intermediate encrypted values during bootstrap.

  Pass a reference to MidBootstrapDecrypter.decrypt to `bootstrap` as the
  `callback` argument.
  """

  def __init__(
      self,
      scheme_params: parameters.SchemeParameters,
      encoding_params: encoding.EncodingParameters,
      lwe_key: lwe.LweSecretKey,
      rlwe_key: rlwe.RlweSecretKey,
  ):
    self.scheme_params = scheme_params
    self.encoding_params = encoding_params
    self.lwe_key = lwe_key
    self.rlwe_key = rlwe_key
    self.starting_cleartext = None

  def _noise_and_bits(self, noise) -> str:
    # A helper to format the noise and its log magnitude. Noise may be zero, in
    # which case treat its log as 1 to avoid domain error.
    abs_bits = math.log2(max(1, abs(noise)))
    return f"{noise} ({abs_bits} bits)"

  def decrypt(
      self,
      name: str,
      value: jnp.ndarray,
      callback_lut: Optional[list[int]] = None,
  ) -> None:
    """Decrypt an intermediate value of bootstrap and report noise growth."""
    lwe_dim = self.scheme_params.lwe_dimension

    if name == "initial":
      ciphertext = value
      plaintext = lwe.decrypt(
          ciphertext,
          self.lwe_key,
          encoding_params=self.encoding_params,
      )
      initial_noise = encoding.extract_noise(
          lwe.decrypt_without_denoising(ciphertext, self.lwe_key),
          self.encoding_params,
      )
      cleartext = encoding.decode(plaintext, self.encoding_params)
      self.starting_cleartext = cleartext
      print(f"Cleartext at bootstrap start = {cleartext}")
      print(
          "Ciphertext noise at bootstrap start = "
          + self._noise_and_bits(initial_noise)
      )

    if name == "approx_ciphertext":
      modulus_switched = value
      log_output_modulus = self.scheme_params.log_mod_degree + 1
      post_mod_switch_key = lwe.LweSecretKey(
          log_modulus=log_output_modulus,
          lwe_dimension=lwe_dim,
          key_data=self.lwe_key.key_data,
      )
      post_mod_switch_encoding_params = encoding.EncodingParameters(
          total_bit_length=log_output_modulus,
          message_bit_length=self.encoding_params.message_bit_length,
          padding_bit_length=self.encoding_params.padding_bit_length,
      )
      plaintext_post_mod_switch = lwe.decrypt_without_denoising(
          modulus_switched, post_mod_switch_key
      )
      cleartext_post_mod_switch = encoding.decode(
          plaintext_post_mod_switch, post_mod_switch_encoding_params
      )
      noise_after = encoding.extract_noise(
          plaintext_post_mod_switch, post_mod_switch_encoding_params
      )
      print(f"Cleartext post modulus switch = {cleartext_post_mod_switch}")
      print(
          "Ciphertext noise post modulus switch = "
          + self._noise_and_bits(noise_after)
      )
      if self.starting_cleartext != cleartext_post_mod_switch:
        print(
            f"DECRYPTION_FAILURE: expected {self.starting_cleartext} "
            f"but got {cleartext_post_mod_switch}"
        )

    if name == "rotated":
      rotated_rlwe_ciphertext = rlwe.RlweCiphertext(
          log_coefficient_modulus=self.rlwe_key.log_coefficient_modulus,
          modulus_degree=self.rlwe_key.modulus_degree,
          message=value,
      )
      rotated_plaintext = rlwe.decrypt(
          ciphertext=rotated_rlwe_ciphertext,
          sk=self.rlwe_key,
          encoding_params=self.encoding_params,
      )
      rotated_decoded = [
          encoding.decode(x, self.encoding_params)
          for x in rotated_plaintext.message
      ]
      print(f"Cleartext rotated polynomial = {rotated_decoded}")

    if name == "extracted":
      extracted = value
      reshaped_key = rlwe.flatten_key(self.rlwe_key)
      plaintext_post_sample_extract = lwe.decrypt_without_denoising(
          extracted, reshaped_key
      )
      cleartext_post_sample_extract = encoding.decode(
          plaintext_post_sample_extract, self.encoding_params
      )
      noise_post_sample_extract = encoding.extract_noise(
          plaintext_post_sample_extract, self.encoding_params
      )
      print(f"Cleartext post sample extract = {cleartext_post_sample_extract}")
      print(
          "Ciphertext noise post sample extract = "
          + self._noise_and_bits(noise_post_sample_extract)
      )

      expected_cleartext = self.starting_cleartext
      if callback_lut:
        expected_cleartext = callback_lut[self.starting_cleartext]
      if expected_cleartext != cleartext_post_sample_extract:
        print(
            f"DECRYPTION_FAILURE: expected {expected_cleartext} "
            f"but got {cleartext_post_sample_extract}. "
        )
        if callback_lut:
          print(
              f"Starting cleartext was {self.starting_cleartext} "
              f"and LUT was {callback_lut}."
          )

    if name == "key_switched":
      key_switched = value
      plaintext_post_key_switch = lwe.decrypt_without_denoising(
          key_switched, self.lwe_key
      )
      cleartext_post_key_switch = encoding.decode(
          plaintext_post_key_switch, self.encoding_params
      )
      noise_post_key_switch = encoding.extract_noise(
          plaintext_post_key_switch, self.encoding_params
      )
      print(f"Cleartext post key switch = {cleartext_post_key_switch}")
      print(
          "Ciphertext noise post key switch = "
          + self._noise_and_bits(noise_post_key_switch)
      )

      expected_cleartext = self.starting_cleartext
      if callback_lut:
        expected_cleartext = callback_lut[self.starting_cleartext]
      if expected_cleartext != cleartext_post_key_switch:
        print(
            f"DECRYPTION_FAILURE: expected {expected_cleartext} "
            f"but got {cleartext_post_key_switch}."
        )
        if callback_lut:
          print(
              f"Starting cleartext was {self.starting_cleartext} "
              f"and LUT was {callback_lut}."
          )



"""API for Boolean FHE operations over the discretized torus."""

import functools
from typing import Any, Callable, Optional

import jax
import jax.numpy as jnp
from jaxite.jaxite_bool import bool_encoding
from jaxite.jaxite_bool import bool_params
from jaxite.jaxite_lib import bootstrap
from jaxite.jaxite_lib import encoding
from jaxite.jaxite_lib import key_switch
from jaxite.jaxite_lib import lwe
from jaxite.jaxite_lib import random_source
from jaxite.jaxite_lib import rgsw
from jaxite.jaxite_lib import rlwe
from jaxite.jaxite_lib import types

Parameters = bool_params.Parameters
ENCODING_PARAMS = bool_encoding.ENCODING_PARAMS
CLEARTEXT_TRUE = bool_encoding.CLEARTEXT_TRUE
CLEARTEXT_FALSE = bool_encoding.CLEARTEXT_FALSE
CLEARTEXT_UNUSED = bool_encoding.CLEARTEXT_UNUSED


class ClientKeySet:
  """A set of secret keys for client use."""

  @property
  def lwe_sk(self) -> lwe.LweSecretKey:
    return self._lwe_sk

  @property
  def rlwe_sk(self) -> rlwe.RlweSecretKey:
    return self._rlwe_sk

  def __init__(
      self,
      params: Parameters,
      lwe_rng: random_source.RandomSource,
      rlwe_rng: random_source.RandomSource,
  ) -> None:
    self._lwe_sk = lwe.gen_key(params.scheme_params, lwe_rng)
    self._rlwe_sk = rlwe.gen_key(params.scheme_params, rlwe_rng)


class ServerKeySet:
  """A set of public keys and relevant parameters for server/cloud use."""

  @property
  def bsk(self) -> bootstrap.BootstrappingKey:
    return self._bsk

  @property
  def ksk(self) -> key_switch.LweKeySwitchingKey:
    return self._ksk

  def __init__(
      self,
      client_key_set: ClientKeySet,
      params: Parameters,
      lwe_rng: random_source.RandomSource,
      rlwe_rng: random_source.RandomSource,
      bootstrap_callback: Optional[Callable[[str, Any], None]] = None,
  ) -> None:
    self.bootstrap_callback = bootstrap_callback
    rgsw_key = rgsw.key_from_rlwe(client_key_set.rlwe_sk)
    self._bsk = bootstrap.gen_bootstrapping_key(
        lwe_sk=client_key_set.lwe_sk,
        rgsw_sk=rgsw_key,
        decomposition_params=params.bs_decomp_params,
        prg=rlwe_rng,
    )

    in_key = rlwe.flatten_key(client_key_set.rlwe_sk)
    self._ksk = key_switch.gen_key(
        decomposition_params=params.ks_decomp_params,
        prg=lwe_rng,
        in_key=in_key,
        out_key=client_key_set.lwe_sk,
    )


def encrypt(
    value: bool, client_key_set: ClientKeySet, prg: random_source.RandomSource
) -> types.LweCiphertext:
  """Encrypts a Boolean value under a given LWE secret key."""
  cleartext = CLEARTEXT_TRUE if value else CLEARTEXT_FALSE
  plaintext = encoding.encode(cleartext, ENCODING_PARAMS)
  return lwe.encrypt(plaintext, client_key_set.lwe_sk, prg)


def decrypt(
    ciphertext: types.LweCiphertext, client_key_set: ClientKeySet
) -> bool:
  """Decrypts an `LweCiphertext` encryption of `True`/`False`."""
  plaintext = lwe.decrypt(ciphertext, client_key_set.lwe_sk, ENCODING_PARAMS)
  cleartext = encoding.decode(plaintext, ENCODING_PARAMS)
  assert cleartext != CLEARTEXT_UNUSED
  return cleartext != 0


def constant(value: bool, params: Parameters) -> types.LweCiphertext:
  return params.noiseless_true if value else params.noiseless_false


def not_(
    ciphertext: types.LweCiphertext, params: Parameters
) -> types.LweCiphertext:
  """Computes NOT of the input ciphertext."""
  return params.noiseless_true - ciphertext


def and_(
    lhs: types.LweCiphertext,
    rhs: types.LweCiphertext,
    server_key_set: ServerKeySet,
    params: Parameters,
) -> types.LweCiphertext:
  """Computes AND of lhs and rhs."""
  return bootstrap.bootstrap(
      2 * lhs + rhs,
      params.lut_poly_by_name('and'),
      server_key_set.bsk,
      server_key_set.ksk,
      params.ks_decomp_params,
      params.bs_decomp_params,
      params.scheme_params,
      callback=server_key_set.bootstrap_callback,
      callback_lut=params.lut_by_name('and').as_cleartext_list,
  )


def andny_(
    lhs: types.LweCiphertext,
    rhs: types.LweCiphertext,
    server_key_set: ServerKeySet,
    params: Parameters,
) -> types.LweCiphertext:
  """Computes ANDNY of lhs andny rhs."""
  return bootstrap.bootstrap(
      2 * lhs + rhs,
      params.lut_poly_by_name('andny'),
      server_key_set.bsk,
      server_key_set.ksk,
      params.ks_decomp_params,
      params.bs_decomp_params,
      params.scheme_params,
      callback=server_key_set.bootstrap_callback,
      callback_lut=params.lut_by_name('andny').as_cleartext_list,
  )


def andyn_(
    lhs: types.LweCiphertext,
    rhs: types.LweCiphertext,
    server_key_set: ServerKeySet,
    params: Parameters,
) -> types.LweCiphertext:
  """Computes ANDYN of lhs andyn rhs."""
  return bootstrap.bootstrap(
      2 * lhs + rhs,
      params.lut_poly_by_name('andyn'),
      server_key_set.bsk,
      server_key_set.ksk,
      params.ks_decomp_params,
      params.bs_decomp_params,
      params.scheme_params,
      callback=server_key_set.bootstrap_callback,
      callback_lut=params.lut_by_name('andyn').as_cleartext_list,
  )


def or_(
    lhs: types.LweCiphertext,
    rhs: types.LweCiphertext,
    server_key_set: ServerKeySet,
    params: Parameters,
) -> types.LweCiphertext:
  """Computes OR of lhs or rhs."""
  return bootstrap.bootstrap(
      2 * lhs + rhs,
      params.lut_poly_by_name('or'),
      server_key_set.bsk,
      server_key_set.ksk,
      params.ks_decomp_params,
      params.bs_decomp_params,
      params.scheme_params,
      callback=server_key_set.bootstrap_callback,
      callback_lut=params.lut_by_name('or').as_cleartext_list,
  )


def orny_(
    lhs: types.LweCiphertext,
    rhs: types.LweCiphertext,
    server_key_set: ServerKeySet,
    params: Parameters,
) -> types.LweCiphertext:
  """Computes ORNY of lhs orny rhs."""
  return bootstrap.bootstrap(
      2 * lhs + rhs,
      params.lut_poly_by_name('orny'),
      server_key_set.bsk,
      server_key_set.ksk,
      params.ks_decomp_params,
      params.bs_decomp_params,
      params.scheme_params,
      callback=server_key_set.bootstrap_callback,
      callback_lut=params.lut_by_name('orny').as_cleartext_list,
  )


def oryn_(
    lhs: types.LweCiphertext,
    rhs: types.LweCiphertext,
    server_key_set: ServerKeySet,
    params: Parameters,
) -> types.LweCiphertext:
  """Computes ORYN of lhs oryn rhs."""
  return bootstrap.bootstrap(
      2 * lhs + rhs,
      params.lut_poly_by_name('oryn'),
      server_key_set.bsk,
      server_key_set.ksk,
      params.ks_decomp_params,
      params.bs_decomp_params,
      params.scheme_params,
      callback=server_key_set.bootstrap_callback,
      callback_lut=params.lut_by_name('oryn').as_cleartext_list,
  )


def nand_(
    lhs: types.LweCiphertext,
    rhs: types.LweCiphertext,
    server_key_set: ServerKeySet,
    params: Parameters,
) -> types.LweCiphertext:
  """Computes NAND of lhs nand rhs."""
  return bootstrap.bootstrap(
      2 * lhs + rhs,
      params.lut_poly_by_name('nand'),
      server_key_set.bsk,
      server_key_set.ksk,
      params.ks_decomp_params,
      params.bs_decomp_params,
      params.scheme_params,
      callback=server_key_set.bootstrap_callback,
      callback_lut=params.lut_by_name('nand').as_cleartext_list,
  )


def nor_(
    lhs: types.LweCiphertext,
    rhs: types.LweCiphertext,
    server_key_set: ServerKeySet,
    params: Parameters,
) -> types.LweCiphertext:
  """Computes NOR of lhs nor rhs."""
  return bootstrap.bootstrap(
      2 * lhs + rhs,
      params.lut_poly_by_name('nor'),
      server_key_set.bsk,
      server_key_set.ksk,
      params.ks_decomp_params,
      params.bs_decomp_params,
      params.scheme_params,
      callback=server_key_set.bootstrap_callback,
      callback_lut=params.lut_by_name('nor').as_cleartext_list,
  )


def xor_(
    lhs: types.LweCiphertext,
    rhs: types.LweCiphertext,
    server_key_set: ServerKeySet,
    params: Parameters,
) -> types.LweCiphertext:
  """Computes XOR of lhs xor rhs."""
  return bootstrap.bootstrap(
      2 * lhs + rhs,
      params.lut_poly_by_name('xor'),
      server_key_set.bsk,
      server_key_set.ksk,
      params.ks_decomp_params,
      params.bs_decomp_params,
      params.scheme_params,
      callback=server_key_set.bootstrap_callback,
      callback_lut=params.lut_by_name('xor').as_cleartext_list,
  )


def xnor_(
    lhs: types.LweCiphertext,
    rhs: types.LweCiphertext,
    server_key_set: ServerKeySet,
    params: Parameters,
) -> types.LweCiphertext:
  """Computes XNOR of lhs xnor rhs."""
  return bootstrap.bootstrap(
      2 * lhs + rhs,
      params.lut_poly_by_name('xnor'),
      server_key_set.bsk,
      server_key_set.ksk,
      params.ks_decomp_params,
      params.bs_decomp_params,
      params.scheme_params,
      callback=server_key_set.bootstrap_callback,
      callback_lut=params.lut_by_name('xnor').as_cleartext_list,
  )


def lut1(
    a: types.LweCiphertext,
    truth_table: int,
    server_key_set: ServerKeySet,
    params: Parameters,
) -> types.LweCiphertext:
  """Computes a 1-LUT."""
  del server_key_set
  if truth_table == 0 or truth_table == 3:
    return constant(bool(truth_table), params)
  if truth_table == 1:
    return not_(a, params)
  if truth_table == 2:
    return a

  raise ValueError(f'Illegal truth table {truth_table} for 1-LUT')


def lut2(
    a: types.LweCiphertext,
    b: types.LweCiphertext,
    truth_table: int,
    server_key_set: ServerKeySet,
    params: Parameters,
) -> types.LweCiphertext:
  """Computes a 2-LUT.

  Over cleartext bits a, b, the operation computed by this function can be
  interpreted as

    truth_table >> {b, a}

  where {b, a} is the unsigned 2-bit integer with bits b, a from most
  significant bit to least significant bit.

  Args:
    a: the least significant bit of the value to lookup
    b: the most significant bit of the value to lookup
    truth_table: the truth table contents
    server_key_set: the FHE server keys
    params: the jaxite_bool parameters

  Returns:
    The LWE ciphertext encrypting the output of the 2-LUT.
  """
  return bootstrap.bootstrap(
      2 * b + a,
      params.lut_poly(num_inputs=2, truth_table=truth_table),
      server_key_set.bsk,
      server_key_set.ksk,
      params.ks_decomp_params,
      params.bs_decomp_params,
      params.scheme_params,
      callback=server_key_set.bootstrap_callback,
      callback_lut=params.lut(
          num_inputs=2, truth_table=truth_table
      ).as_cleartext_list,
  )


def lut3(
    a: types.LweCiphertext,
    b: types.LweCiphertext,
    c: types.LweCiphertext,
    truth_table: int,
    server_key_set: ServerKeySet,
    params: Parameters,
) -> types.LweCiphertext:
  """Computes a 3-LUT.

  Over cleartext bits a, b, c, the operation computed by this function can be
  interpreted as

    truth_table >> {c, b, a}

  where {c, b, a} is the unsigned 3-bit integer with bits c, b, a from most
  significant bit to least-significant bit.

  Args:
    a: the least significant bit of the value to lookup
    b: the middle bit of the value to lookup
    c: the most significant bit of the value to lookup
    truth_table: the truth table contents
    server_key_set: the FHE server keys
    params: the jaxite_bool parameters

  Returns:
    The LWE ciphertext encrypting the output of the 3-LUT.
  """
  return bootstrap.bootstrap(
      4 * c + 2 * b + a,
      params.lut_poly(num_inputs=3, truth_table=truth_table),
      server_key_set.bsk,
      server_key_set.ksk,
      params.ks_decomp_params,
      params.bs_decomp_params,
      params.scheme_params,
      callback=server_key_set.bootstrap_callback,
      callback_lut=params.lut(
          num_inputs=3, truth_table=truth_table
      ).as_cleartext_list,
  )


def cmux_(
    v1: types.LweCiphertext,
    v0: types.LweCiphertext,
    ctrl: types.LweCiphertext,
    server_key_set: ServerKeySet,
    params: Parameters,
) -> types.LweCiphertext:
  truth_table = params.lut_by_name('cmux').truth_table
  return lut3(ctrl, v0, v1, truth_table, server_key_set, params)


@functools.partial(
    jax.pmap,
    in_axes=(0, 0, 0, 0, None, None),
    out_axes=0,
    static_broadcasted_argnums=(4, 5),
)
def pmap_lut3_impl(
    a: types.LweCiphertext,
    b: types.LweCiphertext,
    c: types.LweCiphertext,
    truth_table: jnp.ndarray,
    server_key_set: ServerKeySet,
    params: Parameters,
) -> types.LweCiphertext:
  """A version of lut3 suitable for pmap."""
  return bootstrap.jit_bootstrap(
      4 * c + 2 * b + a,
      truth_table,
      server_key_set.bsk.encrypted_lwe_sk_bits,
      server_key_set.ksk.key_data,
      params.ks_decomp_params,
      params.bs_decomp_params,
      params.scheme_params,
  )


Lut3Args = tuple[jnp.array, jnp.array, jnp.array, jnp.array]


def pmap_lut3(
    inputs: list[Lut3Args], sks: ServerKeySet, params: Parameters
) -> jnp.ndarray:
  """Apply lut3 in parallel across all inputs.

  Args:
    inputs: A list of tuples (a, b, c, truth_table), the corresponding arguments
      to the `jaxite.lut3` function with the same semantics
    sks: the FHE server key set
    params: the jaxite_bool parameters

  Returns:
    The LWE ciphertexts encrypting the output of the 3-LUT,
    organized as rows of a jax array.
  """
  a_inputs = jnp.array([v[0] for v in inputs])
  b_inputs = jnp.array([v[1] for v in inputs])
  c_inputs = jnp.array([v[2] for v in inputs])

  truth_table_cts = jnp.array(
      [params.lut_poly(num_inputs=3, truth_table=v[3]).message for v in inputs]
  )

  return pmap_lut3_impl(
      a_inputs,
      b_inputs,
      c_inputs,
      truth_table_cts,
      sks,
      params,
  )


@functools.partial(
    jax.pmap,
    in_axes=(0, 0, 0, None, None),
    out_axes=0,
    static_broadcasted_argnums=(3, 4),
)
def pmap_lut2_impl(
    a: types.LweCiphertext,
    b: types.LweCiphertext,
    truth_table: jnp.ndarray,
    server_key_set: ServerKeySet,
    params: Parameters,
) -> types.LweCiphertext:
  """A version of lut3 suitable for pmap."""
  return bootstrap.jit_bootstrap(
      2 * b + a,
      truth_table,
      server_key_set.bsk.encrypted_lwe_sk_bits,
      server_key_set.ksk.key_data,
      params.ks_decomp_params,
      params.bs_decomp_params,
      params.scheme_params,
  )


Lut2Args = tuple[jnp.array, jnp.array, jnp.array]


def pmap_lut2(
    inputs: list[Lut2Args], sks: ServerKeySet, params: Parameters
) -> jnp.ndarray:
  """Apply lut2 in parallel across all inputs.

  Args:
    inputs: A list of tuples (a, b, truth_table), the corresponding arguments
      to the `jaxite.lut2` function with the same semantics
    sks: the FHE server key set
    params: the jaxite_bool parameters

  Returns:
    The LWE ciphertexts encrypting the output of the 2-LUT,
    organized as rows of a jax array.
  """
  a_inputs = jnp.array([v[0] for v in inputs])
  b_inputs = jnp.array([v[1] for v in inputs])

  truth_table_cts = jnp.array(
      [params.lut_poly(num_inputs=2, truth_table=v[2]).message for v in inputs]
  )

  return pmap_lut2_impl(
      a_inputs,
      b_inputs,
      truth_table_cts,
      sks,
      params,
  )



"""A module to store hard-coded encoding information about jaxite_bool."""
from jaxite.jaxite_lib import encoding
from jaxite.jaxite_lib import types

ENCODING_PARAMS = encoding.EncodingParameters(
    # leaves 32 - 3 - 1 bits for noise
    total_bit_length=32,
    message_bit_length=3,
    padding_bit_length=1,
)
CLEARTEXT_FALSE = types.LweCleartext(0)
CLEARTEXT_TRUE = types.LweCleartext(1)

# This value is used for the "illegal" entry of a test polynomial, which
# bootstrap should never produce as output if our implementation is correct and
# the ciphertext error does not grow too large. The decryption function has an
# assertion that this value is never the final output of a circuit.
CLEARTEXT_UNUSED = types.LweCleartext(2)



"""A module for operations on test polynomials.

In general, a test polynomial is a degree-N polynomial p(x) in the ring Z/qZ[x]
mod (x^N+1) which represents the lookup table of a single-input function f(x) on
M-bit integers (2^M <= N). The simplest example, and the function used for
non-functional bootstrapping, is the identity function on the message space
Z/2^M Z.

This module constructs test polynomials in such a way as to make them usable for
FHE, in the sense that blind_rotate is used to index a specific coefficient a_j
of the test polynomial. In particular, j = b - <a,s> is the underlying plaintext
of an LWE encrypted message, plus some error term.

The complexity arises because the index j is approximate, and may be negative. A
negative index is interpreted modulo N, but because the ring Z/qZ[x] / (x^N+1)
is negacyclic (i.e., x^N = -1, not 1), the test polynomial p(x) must be
constructed with special care to work with blind rotate. The steps are as
follows:

  1. Start with a raw lookup table as a list L with L[i] = f(i) mod 2^M. This
     table must be length 2^M to account for all possible inputs to the
     function.

  2. Repeat each coefficient some number of times, which corresponds to adding
     protection against approximation error in the input index. The amount of
     repetition per block is N / 2^M.

  3. Rotate the coefficients backwards by half the width of one repetition
     block, to account for the possibility that the input index is negative
     due to error in the blind_rotate input j. I.e., an error-free index input
     to blind_rotate lands in the center of its block.

  4. Encode the coefficients so as to be compatible with the encoding of LWE
     plaintexts, i.e., in such a way that a coefficient that is < 2^M can be
     decoded as if it were an LWE plaintext. This boils down to using the
     padding bits to store test polynomial coefficients that need more than M
     bits. This is required because the output of blind_rotate is interpreted
     as an LWE ciphertext.
"""

import jax.numpy as jnp
from jaxite.jaxite_lib import encoding
from jaxite.jaxite_lib import parameters
from jaxite.jaxite_lib import random_source
from jaxite.jaxite_lib import rlwe


def _pad_for_error(
    polynomial: jnp.ndarray, mod_degree: int, coeff_mod: int
) -> jnp.ndarray:
  """Pads the coefficients of the input polynomial into blocks.

  The curious behavior of this method is that the first block is split across
  the lowest and highest degrees. The test polynomials output by the methods in
  this module are used to "lookup" coefficients by index, but the index input to
  the lookup is scaled and approximate, with symmetric error.

  The width of the block is related to how much error is allowed in the index
  before the wrong value is looked up, and the rotation ensures that symmetric
  error for index 0 (mod polynomial.modulus_degree) will wrap around to the
  highest coefficient and provide the right answer.

  Because the coefficients in the part of the first block wrap around in the
  negative direction (due to the negacyclic property), we must negate those
  terms.

  Args:
    polynomial: the input polynomial coefficients, as a jax array.
    mod_degree: the degree of the resulting polynomial
    coeff_mod: the modulus of the coefficients

  Returns:
    The coefficients padded into blocks of width mod_degree / len(polynomial)
  """
  # We want to fill up the entire output polynomial, which has `mod_degree` many
  # coefficients, with no space left over. Since mod_degree and the message
  # space size are both powers of two, this division is exact and each block has
  # the same size, which is also a power of two.
  block_width = mod_degree // polynomial.shape[0]
  blocks = jnp.repeat(polynomial, block_width)

  split_index = block_width // 2
  first_half_block = blocks[:split_index]
  # Because the coefficient wraps around, we must negate those terms to adhere
  # to the negacyclic property.
  negated_half_block = jnp.mod(-first_half_block, coeff_mod)
  return jnp.concatenate((blocks[split_index:], negated_half_block)).astype(
      jnp.uint32
  )


def gen_test_polynomial(
    cleartext_coefficients: jnp.ndarray,
    encoding_params: encoding.EncodingParameters,
    scheme_params: parameters.SchemeParameters,
) -> rlwe.RlwePlaintext:
  """Generates a test polynomial with padding and encoding.

  The test polynomial generated encodes cleartext_coefficients, and then the
  result is padded so that the total polynomial degree is
  polynomial_modulus_degree - 1.

  Args:
    cleartext_coefficients: the unpadded coefficients of the test polynomial.
    encoding_params: the encoding parameters that define the message space size.
    scheme_params: the scheme parameters that define the polynomial modulus.

  Returns:
    An RlwePlaintext with the appropriate range and padding.

  Raises:
    ValueError: when the `cleartext_coefficients` provided exceeds the desired
    degree of the test polynomial, `mod_degree`.
  """
  mod_degree = scheme_params.polynomial_modulus_degree
  message_space_size = 2**encoding_params.message_bit_length
  if mod_degree < len(cleartext_coefficients):
    raise ValueError(
        f"Degree ({mod_degree}) must exceed the number of "
        f"coefficients ({len(cleartext_coefficients)})."
    )

  if len(cleartext_coefficients) != message_space_size:
    raise ValueError(
        "Coefficients must have length equal to the size of the "
        f"message space, but length was {len(cleartext_coefficients)}"
    )

  padded = _pad_for_error(
      cleartext_coefficients, mod_degree, message_space_size
  )

  # Encode the coefficients, because blind_rotate will produce as output one of
  # the coefficients, which is then interpreted as an LWE ciphertext.
  test_polynomial = encoding.encode(padded, encoding_params)

  return rlwe.RlwePlaintext(
      message=test_polynomial,
      log_coefficient_modulus=scheme_params.log_plaintext_modulus,
      modulus_degree=mod_degree,
  )


def identity_test_polynomial(
    encoding_params: encoding.EncodingParameters,
    scheme_params: parameters.SchemeParameters,
) -> rlwe.RlwePlaintext:
  """Generates a test polynomial for the identity function.

  A helper for gen_test_polynomial which generates the polynomial

    0 + 1x + 2x^2 + ... + (p-1)x^{p-1},

  where p is the size of the message space. This polynomial is then padded and
  adjusted for error appropriately.

  Args:
    encoding_params: the encoding parameters that define the message space size.
    scheme_params: the scheme parameters that define the polynomial modulus.

  Returns:
    An RlwePlaintext representing a padded test polynomial for the identity
    function.
  """
  coeffs = jnp.arange(2**encoding_params.message_bit_length, dtype=jnp.uint32)
  return gen_test_polynomial(
      cleartext_coefficients=coeffs,
      encoding_params=encoding_params,
      scheme_params=scheme_params,
  )


def trivial_encryption(
    test_polynomial: rlwe.RlwePlaintext,
    scheme_params: parameters.SchemeParameters,
) -> rlwe.RlweCiphertext:
  """Trivially encrypts the test polynomial.

  The encryption of the test polynomial is simple for a nuanced reason.  We
  need a valid RLWE encryption with respect to the same secret key used to
  generate the bootstrapping key (the RGSW SK that is used for cmux/external
  product), but that SK is unknown to this function because it is not public.

  However, the "encryption" (0, 0, ..., msg) is a valid RLWE encryption of msg
  with repect to any secret key. This is equivalent to RLWE encryption with a
  PRG that generates zero for every call.

  Args:
    test_polynomial: the test polynomial to encrypt.
    scheme_params: the scheme parameters used for the dimension of the
      encryption.

  Returns:
    An RlweCiphertext with identically zero samples, which encrypts the test
    polynomial with respect to any secret key.
  """

  # We could manually construct the ciphertext here, but for consistency we call
  # out to the real encrypt function with a fake PRG that always outputs zero.
  zero_rng = random_source.ZeroRng()
  empty_sk = rlwe.gen_key(scheme_params, zero_rng)
  return rlwe.encrypt(test_polynomial, empty_sk, zero_rng)


def gen_and_encrypt(
    cleartext_coefficients: jnp.ndarray,
    encoding_params: encoding.EncodingParameters,
    scheme_params: parameters.SchemeParameters,
) -> rlwe.RlweCiphertext:
  """Generates and encrypts a test polynomial for a lookup table.

  Args:
    cleartext_coefficients: the entries of the lookup table.
    encoding_params: the encoding parameters that define the message space size.
    scheme_params: the scheme parameters that define the polynomial modulus.

  Returns:
    An RlweCiphertext representing a padded test polynomial for the given
    look-up table.
  """
  test_poly = gen_test_polynomial(
      cleartext_coefficients, encoding_params, scheme_params
  )
  return trivial_encryption(test_poly, scheme_params)



"""The API for bootstrapping in TFHE."""
import dataclasses
import functools
from typing import Any, Callable, Optional

import jax
import jax.numpy as jnp
from jaxite.jaxite_lib import decomposition
from jaxite.jaxite_lib import key_switch
from jaxite.jaxite_lib import lwe
from jaxite.jaxite_lib import matrix_utils
from jaxite.jaxite_lib import parameters
from jaxite.jaxite_lib import random_source
from jaxite.jaxite_lib import rgsw
from jaxite.jaxite_lib import rlwe
from jaxite.jaxite_lib import types

GEN_BSK_NUM_BATCHES = 15
NON_DIVISIBLE_BATCH_SIZE_WARNING = (
    "Expected lwe_sk_dim to be a multiple of %s, but was %s. "
    "This is OK for tests with small LWE secret key sizes, but may cause "
    "memory exhaustion errors with prod security params. "
    "Proceeding with unbatched BSK keygen."
)


@dataclasses.dataclass
class BootstrappingKey:
  """An array with row j an RGSW encryption of bit j of an LWE secret key."""

  encrypted_lwe_sk_bits: jnp.ndarray


def gen_bootstrapping_key(
    lwe_sk: lwe.LweSecretKey,
    rgsw_sk: rgsw.RgswSecretKey,
    decomposition_params: decomposition.DecompositionParameters,
    prg: random_source.RandomSource,
) -> BootstrappingKey:
  """Generate a bootstrapping key for the given LWE secret key.

  A bootstrapping key is a list of N RGSW ciphertexts, each encrypting a bit
  of a given LWE secret key sk_in.

  Args:
    lwe_sk: the input LWE secret key, to be encrypted by the RGSW key
    rgsw_sk: the RGSW secret key, to use to encrypt the LWE key
    decomposition_params: bit-decomposition parameters needed for RGSW
      encryption
    prg: the random source

  Returns:
    A bootstrapping key.
  """
  k = num_blocks = rgsw_sk.key.rlwe_dimension
  rlwe_sk = rgsw_sk.to_rlwe_secret_key()
  lwe_sk_dim = lwe_sk.key_data.shape[0]
  levels = decomposition_params.level_count
  if lwe_sk_dim % 2 != 0:
    raise ValueError(
        "BMMP bootstrap loop-unrolling technique "
        "requires an even lwe_sk dimension, but got "
        f"{lwe_sk_dim}"
    )
  num_bsk_encryptions = lwe_sk_dim + lwe_sk_dim // 2

  ai_samples = prg.uniform(
      shape=(
          num_bsk_encryptions,
          num_blocks + 1,
          levels,
          k,
          rlwe_sk.modulus_degree,
      ),
      dtype=jnp.uint32,
  )
  error_samples = prg.rounded_normal(
      shape=(num_bsk_encryptions, num_blocks + 1, levels),
      dtype=jnp.uint32,
  )

  # Using the improved blind rotate from Bourse-Minelli-Minihold-Paillier
  # (BMMP17: https://eprint.iacr.org/2017/1114), a trick uses a larger
  # bootstrapping key to reduce the number of external products required by 1/2.
  # Rather than encrypt the secret key bits of the LWE key separately, we
  # encrypt:
  #
  #  BSK_{3i}   = s_{2i} * s_{2i+1},
  #  BSK_{3i+1} = s_{2i} * (1 − s_{2i+1}),
  #  BSK_{3i+2} = (1 − s_{2i}) * s_{2i+1}
  #
  # which enables a bootstrap operation that involves 1/2 as many external
  # products, though this causes the bootstrapping key to be 50% larger.
  lwe_sk_data = lwe_sk.key_data.astype(jnp.uint32)
  bsk_input_pairs = jnp.zeros(num_bsk_encryptions, dtype=jnp.uint32)

  bsk_input_pairs = bsk_input_pairs.at[::3].set(
      jnp.multiply(lwe_sk_data[::2], lwe_sk_data[1::2])
  )
  bsk_input_pairs = bsk_input_pairs.at[1::3].set(
      jnp.multiply(lwe_sk_data[::2], 1 - lwe_sk_data[1::2])
  )
  bsk_input_pairs = bsk_input_pairs.at[2::3].set(
      jnp.multiply(1 - lwe_sk_data[::2], lwe_sk_data[1::2])
  )

  # Applying vmap to the entire jit_encrypt over all sk bits will exhaust the
  # tensor core's memory with prod security parameters. It will eagerly allocate
  # a u32[945,2,8,1,1024,1024] which has size ~60 GiB, but a Dragonfish chip has
  # only 16 GiB. To reduce memory usage, we can use lax.map to batch the
  # computation in 15 batches and then reshape the result at the end. Doing this
  # results in intermediate allocations of size ~4 GiB for prod security
  # parameters, which provides enough space for the remaining allocations of
  # random numbers.
  num_batches = GEN_BSK_NUM_BATCHES
  batch_size = num_bsk_encryptions // num_batches
  if num_bsk_encryptions % num_batches != 0:
    print(NON_DIVISIBLE_BATCH_SIZE_WARNING % (batch_size, num_bsk_encryptions))
    num_batches = 1
    batch_size = num_bsk_encryptions

  ai_samples = ai_samples.reshape((
      num_batches,
      batch_size,
      num_blocks + 1,
      levels,
      k,
      rlwe_sk.modulus_degree,
  ))
  error_samples = error_samples.reshape((
      num_batches,
      batch_size,
      num_blocks + 1,
      levels,
  ))

  bsk_input_pairs = bsk_input_pairs.reshape((num_batches, batch_size))

  vmapped_encrypt = jax.vmap(
      rgsw.jit_encrypt,
      in_axes=(0, None, 0, 0, None, None, None, None),
      out_axes=0,
  )

  def process_one_batch(i):
    ai_samples_batch = ai_samples[i]
    error_samples_batch = error_samples[i]
    input_batch = bsk_input_pairs[i]
    return vmapped_encrypt(
        input_batch,
        rlwe_sk.data,
        ai_samples_batch,
        error_samples_batch,
        decomposition_params,
        num_blocks,
        rlwe_sk.log_coefficient_modulus,
        rlwe_sk.modulus_degree,
    )

  encrypted_lwe_sk_bits = jax.lax.map(
      process_one_batch, jnp.arange(num_batches)
  )
  # The output is stacked by batches, and this reshape combines the first two
  # axes into a single axis of size num_bsk_encryptions.
  encrypted_lwe_sk_bits = encrypted_lwe_sk_bits.reshape((
      num_bsk_encryptions,
      (num_blocks + 1) * levels,
      num_blocks + 1,
      rlwe_sk.modulus_degree,
  ))

  return BootstrappingKey(encrypted_lwe_sk_bits=encrypted_lwe_sk_bits)


@jax.named_call
@functools.partial(jax.jit, static_argnums=(4, 5, 6))
def jit_bootstrap(
    ciphertext: types.LweCiphertext,
    test_poly_ciphertext_message: jnp.ndarray,
    bsk_encrypted_lwe_sk_bits: jnp.ndarray,
    ksk_key_data: jnp.ndarray,
    ks_decomposition_params: decomposition.DecompositionParameters,
    bs_decomposition_params: decomposition.DecompositionParameters,
    scheme_params: parameters.SchemeParameters,
) -> types.LweCiphertext:
  """Apply functional bootstrap to reduce noise in the input ciphertext.

  Args:
    ciphertext: the input LWE ciphertext.
    test_poly_ciphertext_message: The RLWE encrypted polynomial of the cleartext
      message values used to retrieve the target sample during bootstrapping.
    bsk_encrypted_lwe_sk_bits: a bootstrapping key, encrypting the bits of the
      secret key used to encrypt `ciphertext`
    ksk_key_data: a key switching key, required to switch back to the original
      LWE key after bootstrapping switches to an LWE key determined by `bsk`.
    ks_decomposition_params: the decomposition parameters used in the key
      switching key.
    bs_decomposition_params: the decomposition parameters used in the
      bootstrapping key.
    scheme_params: the scheme parameters, used to discretize the ciphertext.

  Returns:
    An encryption of the same underlying message as `ciphertext`, but with
    reduced noise.
  """
  mod_degree = scheme_params.polynomial_modulus_degree
  test_poly_log_coefficient_modulus = scheme_params.log_plaintext_modulus

  approx_ciphertext = lwe.switch_modulus(
      ciphertext,
      log_input_modulus=scheme_params.log_plaintext_modulus,
      log_output_modulus=scheme_params.log_mod_degree + 1,
  )

  # The target LWE sample is contained in the constant term of the output
  # polynomial, which sample_extract extracts as an LWE ciphertext.
  rotated = jit_blind_rotate(
      test_poly_ciphertext_message,
      approx_ciphertext,
      bsk_encrypted_lwe_sk_bits,
      test_poly_log_coefficient_modulus,
      bs_decomposition_params,
  )

  extracted = jit_sample_extract(rotated, mod_degree)

  key_switched = key_switch.jit_switch_key(
      ksk_key_data,
      extracted,
      ks_decomposition_params.level_count,
      ks_decomposition_params.log_base,
  )

  return key_switched


@jax.named_call
def bootstrap(
    ciphertext: types.LweCiphertext,
    test_poly_ciphertext: rlwe.RlweCiphertext,
    bsk: BootstrappingKey,
    ksk: key_switch.LweKeySwitchingKey,
    ks_decomposition_params: decomposition.DecompositionParameters,
    bs_decomposition_params: decomposition.DecompositionParameters,
    scheme_params: parameters.SchemeParameters,
    callback: Optional[Callable[[str, Any], None]] = None,
    **kwargs: Any,
) -> types.LweCiphertext:
  """Apply functional bootstrap to reduce noise in the input ciphertext.

  Args:
    ciphertext: the input LWE ciphertext.
    test_poly_ciphertext: The RLWE encrypted polynomial of the cleartext message
      values used to retrieve the target sample during bootstrapping.
    bsk: a bootstrapping key, encrypting the bits of the secret key used to
      encrypt `ciphertext`.
    ksk: a key switching key, required to switch back to the original LWE key
      after bootstrapping switches to an LWE key determined by `bsk`.
    ks_decomposition_params: the decomposition parameters used in the key
      switching key.
    bs_decomposition_params: the decomposition parameters used in the
      bootstrapping key.
    scheme_params: the scheme parameters, used to discretize the ciphertext.
    callback: an optional callback for tests.
    **kwargs: kwargs forwarded to the callback, if present.

  Returns:
    An encryption of the same underlying message as `ciphertext`, but with
    reduced noise.
  """
  if callback:
    callback("initial", ciphertext, **kwargs)

  mod_degree = scheme_params.polynomial_modulus_degree
  test_poly_log_coefficient_modulus = scheme_params.log_plaintext_modulus

  approx_ciphertext = lwe.switch_modulus(
      ciphertext,
      log_input_modulus=scheme_params.log_plaintext_modulus,
      log_output_modulus=scheme_params.log_mod_degree + 1,
  )
  if callback:
    callback("approx_ciphertext", approx_ciphertext, **kwargs)

  # The target LWE sample is contained in the constant term of the output
  # polynomial, which sample_extract extracts as an LWE ciphertext.
  rotated = jit_blind_rotate(
      test_poly_ciphertext.message,
      approx_ciphertext,
      bsk.encrypted_lwe_sk_bits,
      test_poly_log_coefficient_modulus,
      bs_decomposition_params,
  )
  if callback:
    callback("rotated", rotated, **kwargs)

  extracted = jit_sample_extract(rotated, mod_degree)
  if callback:
    callback("extracted", extracted, **kwargs)

  key_switched = key_switch.jit_switch_key(
      ksk.key_data,
      extracted,
      ks_decomposition_params.level_count,
      ks_decomposition_params.log_base,
  )
  if callback:
    callback("key_switched", key_switched, **kwargs)

  return key_switched


def external_product(
    rgsw_ct: rgsw.RgswCiphertext,
    rlwe_ct: rlwe.RlweCiphertext,
    decomposition_params: decomposition.DecompositionParameters,
) -> rlwe.RlweCiphertext:
  """Compute the external product of an RSGW and RLWE ciphertext."""
  output = jit_external_product(
      rgsw_ct.message, rlwe_ct.message, decomposition_params
  )
  return rlwe.RlweCiphertext(
      log_coefficient_modulus=rlwe_ct.log_coefficient_modulus,
      modulus_degree=rlwe_ct.modulus_degree,
      message=output,
  )


# in_axes = (None, 1) means that the first argument is repeated across all
# calls, while the second argument is mapped across its second index
# (column-wise)
vector_matrix_polymul = jax.jit(
    jax.vmap(matrix_utils.poly_dot_product, in_axes=(None, 1), out_axes=0)
)


@functools.partial(jax.jit, static_argnames="decomposition_params")
def jit_external_product(
    rgsw_ct: jnp.ndarray,
    rlwe_ct: jnp.ndarray,
    decomposition_params: decomposition.DecompositionParameters,
) -> rlwe.RlweCiphertext:
  """Compute the external product of an RSGW and RLWE ciphertext."""
  decomposed_rlwe = decomposition.decompose_rlwe_ciphertext(
      rlwe_ct, decomposition_params
  )
  return vector_matrix_polymul(decomposed_rlwe, rgsw_ct)


def cmux(
    control: rgsw.RgswCiphertext,
    eq_zero: rlwe.RlweCiphertext,
    neq_zero: rlwe.RlweCiphertext,
    decomposition_params: decomposition.DecompositionParameters,
) -> rlwe.RlweCiphertext:
  """Compute CMUX: controlled multiplexer.

  Args:
    control: RGSW ciphertext acting as a selector between eq_zero and neq_zero
    eq_zero: RLWE ciphertext selected if control=0
    neq_zero: RLWE ciphertext selected if control=1
    decomposition_params: decomposition parameters for the external product

  Returns:
    RlwePlaintext: selected RLWE ciphertext
  """
  if eq_zero.coefficient_modulus != neq_zero.coefficient_modulus:
    raise ValueError(
        (
            "Bad params, rlwe ciphertexts must have same coeff modulus"
            " Instead found:\n"
        ),
        f"eq_zero.coeff_modulus={eq_zero.coefficient_modulus}, ",
        f"neq_zero.coeff_modulus={neq_zero.coefficient_modulus}.",
    )
  if eq_zero.modulus_degree != neq_zero.modulus_degree:
    raise ValueError(
        (
            "Bad params, rlwe ciphertexts must have same mod degree."
            " Instead found:\n"
        ),
        f"eq_zero.modulus_degree={eq_zero.modulus_degree}, ",
        f"neq_zero.modulus_degree={neq_zero.modulus_degree}.",
    )
  modulus_degree = eq_zero.modulus_degree
  output = jit_cmux(
      control.message, eq_zero.message, neq_zero.message, decomposition_params
  )
  return rlwe.RlweCiphertext(
      log_coefficient_modulus=eq_zero.log_coefficient_modulus,
      modulus_degree=modulus_degree,
      message=output,
  )


@jax.named_call
@functools.partial(jax.jit, static_argnames="decomposition_params")
def jit_cmux(
    control: jnp.ndarray,
    eq_zero: jnp.ndarray,
    neq_zero: jnp.ndarray,
    decomposition_params: decomposition.DecompositionParameters,
) -> rlwe.RlweCiphertext:
  """A jitted cmux."""
  return (
      eq_zero
      + jit_external_product(  # pytype: disable=bad-return-type  # jax-ndarray
          rgsw_ct=control,
          rlwe_ct=neq_zero - eq_zero,
          decomposition_params=decomposition_params,
      )
  )


def blind_rotate(
    rot_polynomial: rlwe.RlweCiphertext,
    coefficient_index: types.LweCiphertext,
    bsk: BootstrappingKey,
    decomposition_params: decomposition.DecompositionParameters,
) -> rlwe.RlweCiphertext:
  """Rotate an encrypted polynomial `coefficient_index` times.

  Args:
    rot_polynomial: an RLWE encryption of the polynomial P to rotate. This is
      referred to as TGLWE_s'(v) in the TFHE explainer paper
      (https://eprint.iacr.org/2021/1402.pdf). Note that (0, 0, ... v) is a
      valid RLWE encryption of v.
    coefficient_index: the encrypted index J to rotate to the constant term.
      This is referred to as `c^tilde = (a_1^tilde, ... a_n^tilde, b^tilde)`.
    bsk: bootstrapping key, an RGSW encryption of the bits of the secret key `s`
      used to encrypt coefficient_index
    decomposition_params: decomposition parameters for the external product
      operation used in CMUX

  Returns:
    An encryption of X^{-J} * P, which has the J-th coefficient of P as its
    constant term.
  """
  output = jit_blind_rotate(
      rot_polynomial.message,
      coefficient_index,
      bsk.encrypted_lwe_sk_bits,
      rot_polynomial.log_coefficient_modulus,
      decomposition_params,
  )
  return rlwe.RlweCiphertext(
      log_coefficient_modulus=rot_polynomial.log_coefficient_modulus,
      modulus_degree=rot_polynomial.modulus_degree,
      message=output,
  )


@jax.named_call
@functools.partial(jax.jit, static_argnums=(3, 4))
def jit_blind_rotate(
    rot_polynomial: jnp.ndarray,
    coefficient_index: types.LweCiphertext,
    bsk: jnp.ndarray,
    log_coefficient_modulus: int,
    decomposition_params: decomposition.DecompositionParameters,
) -> rlwe.RlweCiphertext:
  """Rotate an encrypted polynomial `coefficient_index` times."""
  # Calculate c' = X^{-b^tilde} * RLWE_s'(v) (for each entry in RLWE_s'(v))
  b_tilde = coefficient_index[coefficient_index.shape[0] - 1]
  c_prime = matrix_utils.monomial_mul_list(
      rot_polynomial,
      -b_tilde,
      log_coefficient_modulus,
  ).astype(jnp.uint32)

  # rot_polynomial is an rlwe ciphertext, so the second dimension determines the
  # degree of the polynomial
  degree = rot_polynomial.shape[1]

  # Using the improved blind rotate from Bourse-Minelli-Minihold-Paillier
  # (BMMP17: https://eprint.iacr.org/2017/1114), a trick uses a larger
  # bootstrapping key to reduce the number of external products required by 1/2.
  num_loop_terms = (coefficient_index.shape[0] - 1) // 2

  def one_bmmp_factor(j):
    power1 = coefficient_index[2 * j] + coefficient_index[2 * j + 1]
    power2 = coefficient_index[2 * j]
    power3 = coefficient_index[2 * j + 1]
    poly_term1 = matrix_utils.x_power_n_minus_1(power1, poly_mod_deg=degree)
    poly_term2 = matrix_utils.x_power_n_minus_1(power2, poly_mod_deg=degree)
    poly_term3 = matrix_utils.x_power_n_minus_1(power3, poly_mod_deg=degree)
    return (
        matrix_utils.poly_mul_const_matrix(poly_term1, bsk[3 * j])
        + matrix_utils.poly_mul_const_matrix(poly_term2, bsk[3 * j + 1])
        + matrix_utils.poly_mul_const_matrix(poly_term3, bsk[3 * j + 2])
    ).astype(jnp.uint32)

  bmmp_factors = jax.vmap(one_bmmp_factor, in_axes=(0,), out_axes=0)(
      jnp.arange(num_loop_terms, dtype=jnp.uint32)
  )

  def one_external_product(j, c_prime_accum):
    return c_prime_accum + jit_external_product(
        rgsw_ct=bmmp_factors[j],
        rlwe_ct=c_prime_accum,
        decomposition_params=decomposition_params,
    )

  return jax.lax.fori_loop(0, num_loop_terms, one_external_product, c_prime)


def sample_extract(ciphertext: rlwe.RlweCiphertext) -> types.LweCiphertext:
  """Extracts an LWE encryption of the constant term encrypted by the input."""
  return jit_sample_extract(ciphertext.message, ciphertext.modulus_degree)


@jax.named_call
@functools.partial(jax.jit, static_argnames="poly_deg")
def jit_sample_extract(
    rlwe_ciphertext: jnp.ndarray, poly_deg: jnp.uint32
) -> types.LweCiphertext:
  """Extracts an LWE encryption of the constant term encrypted by the input.

  Given an RLWE ciphertext with modulus_degree polyN and rlwe_dimension k,
  outputs an LWE ciphertext with lwe_dimension k*polyN.

  Args:
    rlwe_ciphertext: an RLWE encryption of the polynomial to sample extract.
    poly_deg: the polynomial degree of the rlwe_ciphertext.

  Returns:
    An LWE encryption of the constant term of the input polynomial.
  """
  k = jnp.shape(rlwe_ciphertext)[0] - 1  # rlwe_dimension

  # generates an array like [1, -1, -1, -1, ..., -1] of length poly_deg
  # then tiles it vertically rlwe_dimension times.
  extraction_coefficients = jnp.tile(
      jnp.concatenate([
          jnp.ones(1, dtype=jnp.uint32),
          -1 * jnp.ones(poly_deg - 1, dtype=jnp.uint32),
      ]),
      (k, 1),
  )

  # generates an array like [0, poly_deg-1, poly_deg-2, ..., 1] of len poly_deg,
  extraction_indices = jnp.flip(jnp.roll(jnp.arange(poly_deg), poly_deg - 1))

  extracted_sample = (
      extraction_coefficients * rlwe_ciphertext[:-1, extraction_indices]
  )

  b_term_constant_coeff = rlwe_ciphertext[-1][0]

  return jnp.append(
      extracted_sample.reshape((poly_deg * k,)),
      jnp.array([b_term_constant_coeff], dtype=jnp.uint32),
  )



"""Bit-decomposition and recomposition functions."""

import dataclasses
import functools
from typing import NewType
import jax
import jax.numpy as jnp

DecomposedInt = NewType("DecomposedInt", jnp.ndarray)
GadgetMatrix = NewType("GadgetMatrix", jnp.ndarray)
GadgetDecomp = NewType("GadgetDecomp", jnp.ndarray)


@dataclasses.dataclass(frozen=True)
class DecompositionParameters:
  """The parameters to a bit decomposition subroutine.

  Let B = 2^log_base. A bit decomposition computes the first
  level_count digits of the base-B representation of a number,
  which also clears lower-order bits. In the context of TFHE, those lower-order
  bits contain noise and can be safely ignored.
  """

  log_base: int
  level_count: int
  total_bit_length: int = 32


# base_log, num_levels, and total_bit_length are static for the life of the
# program.
@functools.partial(
    jax.jit, static_argnames=("base_log", "num_levels", "total_bit_length")
)
def decompose(
    x: jnp.uint32,
    base_log: int = 4,
    num_levels: int = 3,
    total_bit_length: int = 32,
) -> DecomposedInt:
  """Decompose a number in a given power-of-2 base, up to a given level.

  Let B = 2^base_log. This function computes the first num_levels digits in the
  base-B representation of x, which corresponds to an approximate representation
  of x when the lower-order bits are ignored. In the context of TFHE, those
  lower-order bits contain noise and can be safely ignored.

  Because B is a power of 2, this method partitions a bit string into blocks of
  length base_log, and returns the first num_levels blocks. Here one block
  corresponds to one digit in the base-B representation of x.

  Args:
    x: The number to decompose.
    base_log: The base-2 logarithm of the radix B in which to represent the
      output.
    num_levels: The number of blocks (digits) to return.
    total_bit_length: The bit size of x.

  Returns:
    A DecomposedInt containing the first num_levels digits in the base-B
    representation.
  """
  # Set the most significant base_log many 1's in a bit-string of length
  # total_bit_length, with all other bits unset.  e.g., for base_log=4,
  # mask=0b11110...0
  mask = jnp.uint32((1 << base_log) - 1 << (total_bit_length - base_log))

  # The code that follows is a vectorized version of the following (more
  # legible) loop, in which we iteratively select the next block, shift the
  # resulting bits down to the lowest-order, and shift the mask.
  # Tests were first written with the loop implementation, then refactored
  # with the vectorized version.
  #
  # output = jnp.zeros((num_levels,), dtype=jnp.uint32)
  # for i in range(num_levels):
  #   # Shifts the top-most masked block to the lowest-order bits.
  #   # 1011_0000_0000  ->  0000_0000_1011
  #   digit = (mask & x) >> (total_bit_length - (i + 1) * base_log)
  #   output = jax.ops.index_update(output, i, digit)
  #   mask >>= base_log
  # return output
  output = jnp.repeat(x, num_levels).astype(jnp.uint32)
  masks = mask >> (jnp.arange(num_levels).astype(jnp.uint32) * base_log)
  return (output & masks) >> (
      total_bit_length
      - (1 + jnp.arange(num_levels).astype(jnp.uint32)) * base_log
  )


# base_log, num_levels, and total_bit_length are static for the life of the
# program.
@functools.partial(
    jax.jit, static_argnames=("base_log", "num_levels", "total_bit_length")
)
def recomposition_summands(
    x: jnp.uint32,
    base_log: int = 4,
    num_levels: int = 3,
    total_bit_length: int = 32,
) -> jnp.ndarray:
  """Outputs a list of summands of the digit decomposition.

  Let B = 2^base_log. This function computes the first num_levels digits in the
  base-B representation of x and then scales them with the appropriate power of
  B, so that when the array of recomposition summands are summed, we retrieve an
  approximation of x.

  Because B is a power of 2, this method partitions a bit string into blocks of
  length base_log, and returns the first num_levels blocks. Here one block
  corresponds to one summand in the base-B representation of x.

  Args:
    x: The number to decompose.
    base_log: The base-2 logarithm of the radix B in which to represent the
      output.
    num_levels: The number of blocks (digits) to return.
    total_bit_length: The bit size of x.

  Returns:
    A JAX array containing the first num_levels of summands in the base-B
    representation.
  """
  output = jnp.repeat(x, num_levels)
  return output << (
      total_bit_length
      - (1 + jnp.arange(num_levels, dtype=jnp.uint32)) * base_log
  )


# base_log and total_bit_length are static for the life of the program.
@functools.partial(jax.jit, static_argnames=("base_log", "total_bit_length"))
def recompose(
    digits: DecomposedInt, base_log: int = 4, total_bit_length: int = 32
) -> jnp.uint32:
  """The inverse of decompose.

  Note num_levels == len(digits).

  Args:
    digits: The digits of a bit-decomposed number.
    base_log: The base-2 logarithm of the radix B of the input representation.
    total_bit_length: The bit size of the output integer.

  Returns:
    An integer represented by the DecomposedInt.
  """
  shifted_digits = digits << (
      total_bit_length - (1 + jnp.arange(digits.shape[0])) * base_log
  )
  return jnp.sum(shifted_digits).astype(jnp.uint32)


@functools.partial(
    jax.jit, static_argnames=("base_log", "num_levels", "total_bit_length")
)
def signed_decomposition(
    x: jnp.uint32,
    base_log: jnp.uint32,
    num_levels: int,
    total_bit_length: int = 32,
) -> jnp.ndarray:
  """Compute the signed base-B digit decomposition of a number.

  This method is analogous to decomposition.decompose, however, it restricts
  the digits in the output to the range [-B/2, B/2). This requires extra care
  as follows.

  If one digit in the output is larger than B/2 - 1, then subtracting B from the
  result requires us to add 1 to the next most significant digit. This implies
  we must track a "carry" bit from one digit to the next. Cf.
  b/202561246#comment4 for more details.

  Args:
    x: The number to decompose.
    base_log: The base-2 logarithm of the radix B in which to represent the
      output.
    num_levels: The number of digits to return.
    total_bit_length: The bit size of x.

  Returns:
    A DecomposedInt containing the first num_levels digits in the base-B
    representation.
  """
  # Because a lower order bit might result in a cascading sequence of carry
  # bits, rather than just compute the top num_levels digits, we have to compute
  # all the digits and truncate the return value.
  result = jnp.zeros((total_bit_length // base_log,), dtype=jnp.uint32)
  base = jnp.uint32(1 << base_log)
  digit_mask = base - 1
  base_over_2_threshold = jnp.uint32(1 << (base_log - 1))
  carry = jnp.uint32(0)

  for i in range(len(result)):
    unsigned_digit = ((x >> (i * base_log)) & digit_mask) + carry

    # This mask tests for whether a digit is >= B/2, and hence requires a shift
    # and carry propagation
    carry_mask = unsigned_digit & base_over_2_threshold

    # a tricky way to optionally shift by B. The result is re-cast as a uint32,
    # but this is OK because future arithmetic operations as uint32 will wrap
    # around and treat 2**32-1 properly as -1.
    signed_digit = jnp.uint32(
        jnp.int32(unsigned_digit) - jnp.int32(carry_mask << 1)
    )

    # convert the carry mask to a 0/1 bit to carry to the next digit
    carry = carry_mask >> (base_log - 1)

    result = result.at[i].set(signed_digit)

  # reversed because the digits are computed from least-significant bit to
  # highest-significant bit..
  return result[::-1][:num_levels]


# Applies the signed decomposition to each coefficient of a polynomial
# independently, where the polynomial is represented as a jax array of uints.
# The second and later arguments are the same as `signed_decomposition`, and
# pass through unchanged to each invocation.
signed_decomposition_polynomial = jax.vmap(
    signed_decomposition, in_axes=(0, None, None, None)
)

# Applies the signed decomposition to each coefficient of a list of polynomials
# independently, where one polynomial is represented as a single row of a jax
# array of uints. The second and later arguments are the same as
# `signed_decomposition`, and pass through unchanged to each invocation.
signed_decomposition_polynomial_list = jax.vmap(
    signed_decomposition_polynomial, in_axes=(0, None, None, None)
)


@jax.named_call
@functools.partial(jax.jit, static_argnames="decomposition_params")
def decompose_rlwe_ciphertext(
    rlwe_ct: jnp.ndarray, decomposition_params: DecompositionParameters
) -> jnp.ndarray:
  """Bit-decompose and rearrange an RLWE ciphertext for external product.

  Args:
    rlwe_ct: The underlying message ((k+1, N)-shaped array) of an RLWE
      ciphertext.
    decomposition_params: The DecompositionParameters to use.

  Returns:
    A ((k+1)*L, N) shape ndarray A where A[i, j] is as follows: index i as
    (polynomial, level) from the natural enumeration of range(k+1) x range(L),
    and then A[(polynomial, level), j] is the level'th digit in the signed
    base-L decomposition of coefficient j of the polynomial.
  """
  log_base = decomposition_params.log_base
  level_count = decomposition_params.level_count

  decomposed_polys = signed_decomposition_polynomial_list(
      rlwe_ct, log_base, level_count, decomposition_params.total_bit_length
  )

  # At this point, decomposed_polys has indices in order of (polynomial,
  # coefficient, decomposition level), so we reorder them to (polynomial *
  # decomposition level, coefficient)) and then reshape it so that the rows
  # have dimension (k+1)*L, in major order of polynomial, then decomposition
  # level.
  new_shape = (
      decomposed_polys.shape[0] * decomposed_polys.shape[2],
      decomposed_polys.shape[1],
  )
  return jnp.transpose(decomposed_polys, (0, 2, 1)).reshape(new_shape)


def gadget_matrix(
    decomp_params: DecompositionParameters,
    vector_length: int,
    total_bit_length: int = 32,
) -> GadgetMatrix:
  """Construct a gadget matrix for the decomposition parameters.

  The output matrix is of the following form, where B is the base log of the
  decomposition and l is the number of levels in the decomposition.

  [
   [1/B, 1/B^2, ..., 1/B^l, 0  , 0    , ... , 0    ,        ...        , 0 ],
   [0  ,     ...   , 0    , 1/B, 1/B^2, ... , 1/B^l,        ...        , 0 ],
                                     .
                                        .
                                           .
   [0  ,                ...                 , 0    , 1/B, 1/B^2, ..., 1/B^l]
  ]

  It has the property that (for the same parameters implied),
  inverse_gadget(v).dot(gadget_matrix.T) ~ v, where inverse_gadget(v) is small.
  The approximation is due to the limited number of levels. If B^l ==
  2**total_bit_length, then it is an equality.

  Note, in the Joye paper the vector_length is defined as "n+1" because it
  corresponds to an (R)LWE encryption of size n, and the extra "1" corresponds
  to the last entry `b=sum(a_i s_i) + message + noise`.

  Args:
    decomp_params: The DecompositionParameters to use.
    vector_length: The length of the output vector when the gadget matrix is
      applied to it. If this is M, then the gadget matrix will have shape (M,
      (M*decomp_params.level_count)).
    total_bit_length: The total bit size of the input integers.

  Returns:
    The gadget matrix, in T_q^{(n+1) x (n+1)L}
  """
  base = 2**decomp_params.log_base
  levels = decomp_params.level_count

  if 2**total_bit_length % (base**levels) != 0:
    raise ValueError(
        "Bad params. base**levels must divide total_bit_length. "
        f"Instead found base={base}, levels={levels}, "
        f"total_bit_length={total_bit_length}"
    )

  powers = (jnp.float32(1 / base) ** jnp.arange(1, levels + 1)).reshape(
      (levels, 1)
  )
  return jnp.kron(jnp.identity((vector_length)), powers).T


def inverse_gadget(
    vector: jnp.array,
    decomp_params: DecompositionParameters,
    total_bit_length: int = 32,
) -> GadgetDecomp:
  """Compute the inverse gadget decomposition for a vector.

  Args:
    vector: The vector to apply, with entries in [-1/2, 1/2)
    decomp_params: The DecompositionParameters to use
    total_bit_length: The total bit size of the input integers.

  Returns:
    A flattened vector, the output of the inverse gadget operation.
  """
  log_base = decomp_params.log_base
  base = 2**log_base
  levels = decomp_params.level_count

  scaled = vector * jnp.float32(base) ** levels
  scaled_rounded = jnp.rint(scaled).astype(jnp.uint32)
  return jnp.array(
      [decompose(x, log_base, levels, total_bit_length) for x in scaled_rounded]
  ).flatten()



"""A helper module for constructing look-up tables (LUTs)."""
import functools
import itertools
from typing import Callable

import jax.numpy as jnp
from jaxite.jaxite_bool import bool_encoding
from jaxite.jaxite_lib import parameters
from jaxite.jaxite_lib import rlwe
from jaxite.jaxite_lib import test_polynomial
from jaxite.jaxite_lib import types


class LookUpTable:
  """A representation of a truth table.

  Attrs:
    num_inputs: the number of inputs to the boolean function represented.
    truth_table: an integer containing the truth table values, where the LSB is
      the output for the input "all false", the MSB is the output for the input
      "all true".
  """

  def __init__(self, num_inputs: int, truth_table: int):
    self.num_inputs = num_inputs
    self.truth_table = truth_table

  @functools.cached_property
  def _values_from_lsb_to_msb(self) -> tuple[int, ...]:
    return tuple(
        ((self.truth_table & (1 << i)) >> i)
        for i in range(2**self.num_inputs)
    )

  @functools.cached_property
  def as_cleartext_list(self) -> list[types.LweCleartext]:
    """Return the truth table as a list of LWE cleartexts."""
    return [
        bool_encoding.CLEARTEXT_TRUE if bit else bool_encoding.CLEARTEXT_FALSE
        for bit in self._values_from_lsb_to_msb
    ]

  @functools.lru_cache
  def as_rlwe_test_polynomial(
      self,
      scheme_params: parameters.SchemeParameters,
  ) -> rlwe.RlweCiphertext:
    """Return the truth table as an encrypted RLWE test polynomial."""
    msg_bits = bool_encoding.ENCODING_PARAMS.message_bit_length
    padding_required = 2**msg_bits - len(self.as_cleartext_list)

    if padding_required < 0:
      raise ValueError(
          f'Mismatch between LUT {self.truth_table} and '
          f'encoding params {bool_encoding.ENCODING_PARAMS}'
      )

    padded_coefficients = self.as_cleartext_list + (
        [bool_encoding.CLEARTEXT_UNUSED] * padding_required
    )
    test_poly_coefficients = jnp.array(padded_coefficients, dtype=jnp.uint32)
    return test_polynomial.gen_and_encrypt(
        test_poly_coefficients,
        bool_encoding.ENCODING_PARAMS,
        scheme_params,
    )

  def __str__(self) -> str:
    row_fmt = '{input} -> {value}'
    return '\n'.join(
        row_fmt.format(
            input=bin(input)[2:].rjust(self.num_inputs, '0'), value=bit
        )
        for (input, bit) in enumerate(self._values_from_lsb_to_msb)
    )

  def __repr__(self) -> str:
    return (
        'LookUpTable('
        f'num_inputs={self.num_inputs}, '
        f'truth_table={self.truth_table}, '
        ')'
    )

  def __hash__(self) -> int:
    return hash((self.num_inputs, self.truth_table))


def from_callable(
    num_inputs: int,
    # no way to type-annotate a Callable with any number of bool args
    fn: Callable[..., bool],
) -> LookUpTable:
  """Construct a LookUpTable from a callable of bools."""
  inputs = itertools.product([False, True], repeat=num_inputs)
  truth_table = 0
  bit = 0
  for input_bits in inputs:
    truth_table |= int(fn(*input_bits)) << bit
    bit += 1
  return LookUpTable(num_inputs=num_inputs, truth_table=truth_table)


FUNC_NAME_TO_LUT2 = {
    'and': from_callable(2, lambda x, y: x and y),
    'andny': from_callable(2, lambda x, y: not x and y),
    'andyn': from_callable(2, lambda x, y: x and not y),
    'nand': from_callable(2, lambda x, y: not (x and y)),
    'nor': from_callable(2, lambda x, y: not (x or y)),
    'or': from_callable(2, lambda x, y: x or y),
    'orny': from_callable(2, lambda x, y: not x or y),
    'oryn': from_callable(2, lambda x, y: x or not y),
    'xnor': from_callable(2, lambda x, y: x == y),
    'xor': from_callable(2, lambda x, y: x != y),
    'cmux': from_callable(3, lambda v1, v0, ctrl: v1 if ctrl else v0),
}


class LutCache:
  """A cache for LUT test polynomials for chosen scheme parameters."""

  def __init__(self, scheme_params: parameters.SchemeParameters):
    self.scheme_params = scheme_params
    self.test_poly_cache: dict[int, LookUpTable] = {}

  @functools.lru_cache
  def lut(self, num_inputs: int, lut_as_int: int) -> LookUpTable:
    self.test_poly_cache[lut_as_int] = LookUpTable(
        num_inputs=num_inputs, truth_table=lut_as_int
    )
    return self.test_poly_cache[lut_as_int]

  def lut_by_name(self, name: str) -> LookUpTable:
    if name not in FUNC_NAME_TO_LUT2:
      raise ValueError(
          f'Unknown gate name "{name}". Use `get_lut` or else use a '
          f'named 2-bit gate from {list(FUNC_NAME_TO_LUT2.keys())}'
      )
    key = FUNC_NAME_TO_LUT2[name].truth_table
    return self.lut(num_inputs=2, lut_as_int=key)

  def lut_poly(self, num_inputs: int, lut_as_int: int) -> rlwe.RlweCiphertext:
    return self.lut(num_inputs, lut_as_int).as_rlwe_test_polynomial(
        self.scheme_params
    )

  def lut_poly_by_name(self, name: str) -> rlwe.RlweCiphertext:
    """A helper for users that want to query the gate by name."""
    return self.lut_by_name(name).as_rlwe_test_polynomial(self.scheme_params)



"""LWE encryption scheme."""

import dataclasses
import functools

import jax
import jax.numpy as jnp
from jaxite.jaxite_lib import encoding
from jaxite.jaxite_lib import matrix_utils
from jaxite.jaxite_lib import parameters
from jaxite.jaxite_lib import random_source
from jaxite.jaxite_lib import types


@dataclasses.dataclass
class LweSecretKey:
  """A secret key for the LWE encryption scheme."""

  # the q in Z/2^qZ
  log_modulus: int

  # the length of the sampled key_data vector s,
  # equal to len(LweCiphertext) - 1.
  # Encryption involves sampling a vector a = (a_1, ..., a_{lwe_dimension})
  # and computing a dot product <a, s> + error.
  lwe_dimension: int

  # the binary values (s_1, ..., s_{lwe_dimension})
  # used as a dot product multiplicand when encrypting.
  key_data: jnp.ndarray

  @property
  def modulus(self) -> jnp.uint32:
    return 2 ** jnp.uint32(self.log_modulus)


def gen_key(
    params: parameters.SchemeParameters, prg: random_source.RandomSource
) -> LweSecretKey:
  """Generate an LWE secret key."""
  return LweSecretKey(
      log_modulus=params.log_plaintext_modulus,
      lwe_dimension=params.lwe_dimension,
      key_data=prg.sk_uniform(shape=(params.lwe_dimension,), dtype=jnp.uint32),
  )


def encrypt(
    plaintext: types.LwePlaintext,
    sk: LweSecretKey,
    prg: random_source.RandomSource,
) -> types.LweCiphertext:
  """Encrypt an LWE plaintext."""
  ai_samples = prg.uniform(shape=(sk.lwe_dimension,), dtype=jnp.uint32)
  error_sample = prg.rounded_normal()
  return jit_encrypt(
      plaintext,
      sk.key_data,
      ai_samples,
      error_sample,
      sk.log_modulus,
  )


@functools.partial(jax.jit, static_argnames="log_modulus")
def jit_encrypt(
    plaintext: types.LwePlaintext,
    key_data: jnp.ndarray,
    ai_samples: jnp.ndarray,
    error_sample: jnp.uint32,
    log_modulus: int,
) -> types.LweCiphertext:
  """Encrypt an LWE plaintext with pre-computed randomness."""
  clean_product = jnp.dot(ai_samples, key_data) + plaintext
  obfuscated_product = clean_product + error_sample
  if log_modulus < 32:
    obfuscated_product = obfuscated_product % (jnp.uint32(2) ** log_modulus)
  return jnp.append(ai_samples, obfuscated_product)


def decrypt_without_denoising(
    ciphertext: types.LweCiphertext,
    sk: LweSecretKey,
) -> types.LwePlaintext:
  """Decrypt an LWE ciphertext without removing noise."""
  obfuscated_plaintext = ciphertext[-1] - jnp.dot(ciphertext[:-1], sk.key_data)
  if sk.log_modulus < 32:
    obfuscated_plaintext = jnp.mod(obfuscated_plaintext, sk.modulus)

  return obfuscated_plaintext


def decrypt(
    ciphertext: types.LweCiphertext,
    sk: LweSecretKey,
    encoding_params: encoding.EncodingParameters,
) -> types.LweCleartext:
  """Decrypt and remove the noise from an LWE ciphertext."""
  return encoding.remove_noise(  # pytype: disable=bad-return-type  # jax-ndarray
      decrypt_without_denoising(ciphertext, sk), encoding_params
  )


def noiseless_embedding(
    plaintext: types.LwePlaintext, lwe_dimension: int
) -> types.LweCiphertext:
  """Returns a noiseless LweCiphertext embedding of `plaintext`."""
  samples = jnp.zeros((lwe_dimension,), dtype=jnp.uint32)
  return jnp.append(samples, plaintext)


@jax.jit
def switch_modulus(
    ciphertext: types.LweCiphertext,
    log_input_modulus: jnp.uint32,
    log_output_modulus: jnp.uint32,
) -> types.LweCiphertext:
  """Perform a modulus switch on the input ciphertext.

  The input and output moduli must be powers of two, and they're passed in as
  their logarithms.

  The extra error introduced by modulus switching is at most 0.5 * (lwe_dim+1),
  (note that error is interpreted in the new modulus), and the total error after
  modulus switching is at most

    e_input + 0.5 (lwe_dim + 1),

  where e_input is the error present in `ciphertext`.

  Args:
    ciphertext: the LWE ciphertext to modulus switch
    log_input_modulus: the log2 of the modulus to switch from
    log_output_modulus: the log2 of the modulus to switch to

  Returns:
    An LWE cipherext of the same underlying message, but with values expressed
    with respect to 2**log_output_modulus.
  """
  divisor = jnp.uint32(2) ** (log_input_modulus - log_output_modulus)
  scaled_ciphertext = matrix_utils.integer_div(ciphertext, divisor)
  approx_ciphertext = jnp.mod(
      scaled_ciphertext, jnp.uint32(2) ** log_output_modulus
  )
  return approx_ciphertext



"""LWE Key-switching utils."""

import dataclasses
import functools

import jax
import jax.numpy as jnp
from jaxite.jaxite_lib import decomposition
from jaxite.jaxite_lib import lwe
from jaxite.jaxite_lib import random_source
from jaxite.jaxite_lib import types


@dataclasses.dataclass
class LweKeySwitchingKey:
  """A public key used to switch keys encrypted an LWE ciphertext."""

  # the q in Z/qZ, same as the LwePlaintext space
  modulus: jnp.uint32

  # the length of the sampled key_data vector s,
  # equal to len(LweCiphertext) - 1.
  # Encryption involves sampling a vector a = (a_1, ..., a_{lwe_dimension})
  # and computing a dot product <a, s> + error.
  lwe_dimension: int

  # the decomposition log base
  decomposition_log_base: int

  # the decomposition levels
  decomposition_level_count: int

  # size of switch_keys output plus one
  lwe_size: int

  # the key switching key data, with shape
  # (in_key.lwe_dimension, decomposition_level, out_key.lwe_dimension + 1)
  # note that for TFHE, the out_key.lwe_dimension = N * in_key.lwe_dimension,
  # where N is the polynomial modulus degree in the RLWE/RGSW parameters.
  key_data: jnp.ndarray


def gen_key(
    decomposition_params: decomposition.DecompositionParameters,
    prg: random_source.RandomSource,
    in_key: lwe.LweSecretKey,
    out_key: lwe.LweSecretKey,
) -> LweKeySwitchingKey:
  """Generate an LWE key switching key."""
  num_levels = decomposition_params.level_count
  n = in_key.lwe_dimension
  lwe_size = out_key.lwe_dimension + 1

  # For each bit of the input key (size n), decompose into num_levels pieces and
  # encrypt under the out_key (encryptions have size out_key.lwe_dimension + 1)
  #
  # in_key: length n
  # out_key: of length m
  #
  # "Powers of 2" each term of the in_key (n of these) into t levels, encrypt
  # under out_key to produce a m + 1 length vector
  #
  # Each index corresponds to the decomposition and encryption of one of the
  # bits of the input key.
  # The resulting shape of the key switching key is (n, num_levels, lwe_size)
  lwe_ai_samples = prg.uniform(
      shape=(n, num_levels, out_key.lwe_dimension),
      dtype=jnp.uint32,
  )
  lwe_error_samples = prg.rounded_normal(
      shape=(n, num_levels), dtype=jnp.uint32
  )
  key_data = jnp.zeros((n, num_levels, lwe_size), dtype=jnp.uint32)

  vmap_lwe_encrypt = jax.vmap(
      lwe.jit_encrypt, in_axes=(0, None, 0, 0, None), out_axes=0
  )

  def decompose_and_encrypt(in_key_bit, lwe_ai_samples, lwe_error_samples):
    """Decompose one bit of the input key, then encrypt each term."""
    powers_of_b_pts = decomposition.recomposition_summands(
        in_key_bit,
        base_log=decomposition_params.log_base,
        num_levels=num_levels,
    )
    return vmap_lwe_encrypt(
        powers_of_b_pts,
        out_key.key_data,
        lwe_ai_samples,
        lwe_error_samples,
        out_key.log_modulus,
    )

  key_data = jax.vmap(
      decompose_and_encrypt,
      in_axes=(0, 0, 0),
      out_axes=0,
  )(in_key.key_data, lwe_ai_samples, lwe_error_samples)

  return LweKeySwitchingKey(
      modulus=jnp.uint32(out_key.modulus),
      lwe_dimension=out_key.lwe_dimension,
      decomposition_log_base=decomposition_params.log_base,
      decomposition_level_count=num_levels,
      lwe_size=lwe_size,
      key_data=key_data,
  )


def switch_key(
    ksk: LweKeySwitchingKey, inp: types.LweCiphertext
) -> types.LweCiphertext:
  """Perform the key switch operation on an LWE ciphertext."""
  return jit_switch_key(
      ksk.key_data,
      inp,
      ksk.decomposition_level_count,
      ksk.decomposition_log_base,
  )


@jax.named_call
@functools.partial(jax.jit, static_argnames=("num_levels", "base_log"))
def jit_switch_key(
    ksk: jnp.ndarray,
    inp: types.LweCiphertext,
    num_levels: int,
    base_log: int,
) -> types.LweCiphertext:
  """Perform the key switch operation on an LWE ciphertext."""
  # Decompose the inp ciphertext into num_levels x lwe_size
  output_shape = jnp.shape(ksk[0][0])
  output_data = jnp.zeros(output_shape, dtype=jnp.uint32)
  # Set the b' value.
  output_data = output_data.at[-1].set(inp[-1])

  def decomp_dot(ksk_i, inp_i):
    decomp_i = decomposition.signed_decomposition(inp_i, base_log, num_levels)
    return jnp.dot(decomp_i, ksk_i)

  sum_data = jnp.sum(
      jax.vmap(decomp_dot, in_axes=(0, 0), out_axes=1)(ksk, inp[:-1]), axis=1
  )

  return output_data - sum_data



"""A module defining sources of randomness for TFHE."""

import abc
from collections.abc import Sequence
import functools
import itertools
import operator
import random
from typing import Any, Callable

import jax
import jax.numpy as jnp


def _shape_generator(
    generator: Callable[[], Any],
    shape: Sequence[int],
    dtype: jnp.dtype = jnp.uint32,
) -> jax.Array:
  """Generates an array of data generated by a given function.

  Utility function to generate a JAX ndarray of values given by repeatedly
  invoking the passed in generator function.  Used for RNGs that run on CPU
  serially.

  Args:
    generator: A parameterless function that returns data
    shape: The shape of the JAX ndarray to generate
    dtype: The data type of the JAX ndarray to generate

  Returns:
    A JAX ndarray of the given shape, using data generated by repeatedly calling
    the given generator function.  The data is cast to the given dtype if
    possible.
  """
  if any(d < 0 for d in shape):
    raise ValueError(f"Shape dimensions must be positive: got {shape}.")
  num_elements = functools.reduce(operator.mul, shape, 1)
  return jnp.reshape(
      jnp.array([generator() for _ in range(num_elements)], dtype), shape
  )


class RandomSource(abc.ABC):
  """An interface for a source of randomness."""

  @abc.abstractmethod
  def uniform(
      self, shape: Sequence[int] = (1,), dtype: jnp.dtype = jnp.uint32
  ) -> jax.Array:
    raise NotImplementedError("Subclasses *must* implement `uniform`.")

  @abc.abstractmethod
  def rounded_normal(
      self, shape: Sequence[int] = (1,), dtype: jnp.dtype = jnp.uint32
  ) -> jax.Array:
    raise NotImplementedError("Subclasses *must* implement `rounded_normal`.")

  @abc.abstractmethod
  def sk_uniform(
      self, shape: Sequence[int] = (1,), dtype: jnp.dtype = jnp.uint32
  ) -> jax.Array:
    """Generates random numbers in {0, 1}, for use in generating secret keys."""
    raise NotImplementedError("Subclasses *must* implement `sk_uniform`.")


class CycleRng(RandomSource):
  """An RNG that cycles through a deterministic set of values."""

  def __init__(self, const_normal_noise: int = 0) -> None:
    self.const_normal_noise = const_normal_noise
    # This results in `uniform` having binary-valued outputs
    self.uniform_data = tuple(map(int, bin(13**39**2 // 53)[2:]))
    self.noise = iter(itertools.cycle(self.uniform_data))

  def uniform(
      self, shape: Sequence[int] = (1,), dtype: jnp.dtype = jnp.uint32
  ) -> jax.Array:
    return _shape_generator(lambda: next(self.noise), shape, dtype)

  def sk_uniform(
      self, shape: Sequence[int] = (1,), dtype: jnp.dtype = jnp.uint32
  ) -> jax.Array:
    return self.uniform(shape, dtype=dtype)

  def rounded_normal(
      self, shape: Sequence[int] = (1,), dtype: jnp.dtype = jnp.uint32
  ) -> jax.Array:
    return _shape_generator(lambda: self.const_normal_noise, shape, dtype)


class SystemRandomSource(RandomSource):
  """A cryptographically secure random source that runs on CPU serially.

  Implemented as a wrapper on top of random.SystemRandom().
  """

  def __init__(
      self,
      uniform_bounds: tuple[int, int] = (0, 2**32 - 1),
      normal_std: int = 1,
  ) -> None:
    self.uniform_bounds = uniform_bounds
    self.normal_std = normal_std
    self.rng = random.SystemRandom()

  def uniform(
      self, shape: Sequence[int] = (1,), dtype: jnp.dtype = jnp.uint32
  ) -> jax.Array:
    return _shape_generator(
        lambda: self.rng.randint(*self.uniform_bounds), shape, dtype
    )

  def rounded_normal(
      self, shape: Sequence[int] = (1,), dtype: jnp.dtype = jnp.uint32
  ) -> jax.Array:
    return _shape_generator(
        lambda: round(self.rng.normalvariate(0, self.normal_std)), shape, dtype
    )

  def sk_uniform(
      self, shape: Sequence[int] = (1,), dtype: jnp.dtype = jnp.uint32
  ) -> jax.Array:
    return _shape_generator(lambda: self.rng.randint(0, 1), shape, dtype)


class PseudorandomSource(RandomSource):
  """An insecure random source based on the Python stdlib."""

  def __init__(
      self,
      uniform_bounds: tuple[int, int] = (0, 2**32 - 1),
      normal_std: int = 1,
      seed: int = 1,
  ) -> None:
    self.uniform_bounds = uniform_bounds
    self.normal_std = normal_std
    self.rng = random.Random(seed)

  def uniform(
      self, shape: Sequence[int] = (1,), dtype: jnp.dtype = jnp.uint32
  ) -> jax.Array:
    return _shape_generator(
        lambda: self.rng.randint(*self.uniform_bounds), shape, dtype
    )

  def rounded_normal(
      self, shape: Sequence[int] = (1,), dtype: jnp.dtype = jnp.uint32
  ) -> jax.Array:
    return _shape_generator(
        lambda: round(self.rng.normalvariate(0, self.normal_std)), shape, dtype
    )

  def sk_uniform(
      self, shape: Sequence[int] = (1,), dtype: jnp.dtype = jnp.uint32
  ) -> jax.Array:
    return _shape_generator(lambda: self.rng.randint(0, 1), shape, dtype)


class NormalOnlyRng(RandomSource):
  """A test-only RNG.

  Uniform returns zeros, rounded_normal returns real RNG samples.
  """

  def __init__(self, normal_std: int = 1, seed: int = 1) -> None:
    self.normal_std = normal_std
    self.rng = random.Random(seed)

  def uniform(
      self, shape: Sequence[int] = (1,), dtype: jnp.dtype = jnp.uint32
  ) -> jax.Array:
    return jnp.zeros(shape, dtype=dtype)

  def rounded_normal(
      self, shape: Sequence[int] = (1,), dtype: jnp.dtype = jnp.uint32
  ) -> jax.Array:
    return _shape_generator(
        lambda: round(self.rng.normalvariate(0, self.normal_std)), shape, dtype
    )

  def sk_uniform(
      self, shape: Sequence[int] = (1,), dtype: jnp.dtype = jnp.uint32
  ) -> jax.Array:
    return _shape_generator(lambda: self.rng.randint(0, 1), shape, dtype)


class ConstantUniformRng(RandomSource):
  """A test-only RNG that sets the output of uniform() to a fixed value."""

  def __init__(
      self, const_uniform: int = 1, normal_std: int = 1, seed: int = 1
  ) -> None:
    self.const_uniform = const_uniform
    self.normal_std = normal_std
    self.rng = random.Random(seed)

  def uniform(
      self, shape: Sequence[int] = (1,), dtype: jnp.dtype = jnp.uint32
  ) -> jax.Array:
    return jnp.ones(shape, dtype=dtype) * self.const_uniform

  def rounded_normal(
      self, shape: Sequence[int] = (1,), dtype: jnp.dtype = jnp.uint32
  ) -> jax.Array:
    return _shape_generator(
        lambda: round(self.rng.normalvariate(0, self.normal_std)), shape, dtype
    )

  def sk_uniform(
      self, shape: Sequence[int] = (1,), dtype: jnp.dtype = jnp.uint32
  ) -> jax.Array:
    return _shape_generator(lambda: self.rng.randint(0, 1), shape, dtype)


class ZeroRng(RandomSource):
  """An random generator that always outputs zero."""

  def uniform(
      self, shape: Sequence[int] = (1,), dtype: jnp.dtype = jnp.uint32
  ) -> jax.Array:
    return jnp.zeros(shape, dtype=dtype)

  def rounded_normal(
      self, shape: Sequence[int] = (1,), dtype: jnp.dtype = jnp.uint32
  ) -> jax.Array:
    return jnp.zeros(shape, dtype=dtype)

  def sk_uniform(
      self, shape: Sequence[int] = (1,), dtype: jnp.dtype = jnp.uint32
  ) -> jax.Array:
    return jnp.zeros(shape, dtype=dtype)


_varying_params = (
    ((0, 1), 0),
    ((0, 1), 1),
    ((0, 2**5), 1),
    ((0, 2**15), 1),
    ((0, 2**30), 1),
    ((0, 2**32 - 1), 1),
    ((0, 2**5), 2**5),
    ((0, 2**15), 2**5),
    ((0, 2**30), 2**5),
    ((0, 2**32 - 1), 2**5),
    ((0, 2**32 - 1), 2**7),
)
VARYING_MAGNITUDE_TEST_RNGS = [
    PseudorandomSource(uniform_bounds=uniform_bounds, normal_std=std, seed=1)
    for (uniform_bounds, std) in _varying_params
]

ALL_RNGS = [
    ConstantUniformRng,
    CycleRng,
    NormalOnlyRng,
    PseudorandomSource,
    SystemRandomSource,
    ZeroRng,
]



"""Class encapsulating params for TFHE."""

import dataclasses
import math


@dataclasses.dataclass(frozen=True)
class SchemeParameters:
  """Scheme parameters for TFHE."""

  # The dimension of the LWE secret key vector.
  # Note an encryption is (a_1, a_2, ..., a_n, b),
  # so an LweCiphertext has length lwe_dimension + 1
  lwe_dimension: int

  # the modulus to use for the LWE plaintext
  plaintext_modulus: int

  # dimension of the RLWE ciphertext
  # An RLWE encryption is composed of (a_1, ..., a_n, b),
  # where n = rlwe_dimension.
  rlwe_dimension: int

  # dimension of N in the polynomial modulus x^N + 1
  polynomial_modulus_degree: int

  # the log of polynomial_modulus_degree
  log_mod_degree: int = dataclasses.field(init=False)

  # the log of polynomial_modulus_degree
  log_plaintext_modulus: int = dataclasses.field(init=False)

  def __post_init__(self) -> None:
    object.__setattr__(
        self,
        'log_mod_degree',
        int(round(math.log2(self.polynomial_modulus_degree))),
    )
    object.__setattr__(
        self,
        'log_plaintext_modulus',
        int(round(math.log2(self.plaintext_modulus))),
    )



"""RLWE encryption scheme."""

import dataclasses
import functools

import jax
import jax.numpy as jnp
from jaxite.jaxite_lib import encoding
from jaxite.jaxite_lib import lwe
from jaxite.jaxite_lib import matrix_utils
from jaxite.jaxite_lib import parameters
from jaxite.jaxite_lib import random_source


@dataclasses.dataclass
class RlwePlaintext:
  """An RLWE plaintext is a polynomial in (Z/qZ)[X] / (X^N + 1)."""

  # the log of the modulus q of the polynomial coefficients
  log_coefficient_modulus: int

  # the degree N of the ring modulus polynomial.
  modulus_degree: int

  # the coefficients of the polynomial,
  # starting from lowest degree to highest.
  message: jnp.ndarray

  @property
  def coefficient_modulus(self) -> jnp.uint32:
    return 2 ** jnp.uint32(self.log_coefficient_modulus)

  def __str__(self) -> str:
    # this does not need to be fast because it will only be used in development.
    s = ' + '.join(
        f'{coeff} x^{power}'
        for (power, coeff) in enumerate(self.message)
        if coeff != 0
    )
    return s if s else '0'


@dataclasses.dataclass
class RlweCiphertext:
  """An RLWE ciphertext.

  An RLWE ciphertext is a list of k+1 polynomials in (Z/qZ)[X] / (X^N + 1),
  where the last polynomial is a dot product with the secret key, plus noise.
  """

  # the log of the modulus q of the polynomial coefficients
  log_coefficient_modulus: int

  # the degree N of the ring modulus polynomial.
  modulus_degree: int

  # the polynomials, packed so each row corresponds to one polynomial, and each
  # column corresponds to a coefficient of the same degree.  The first colum is
  # the lowest degree.
  message: jnp.ndarray

  @property
  def coefficient_modulus(self) -> jnp.uint32:
    return 2 ** jnp.uint32(self.log_coefficient_modulus)

  def _single_row_str(self, row):
    return str(
        RlwePlaintext(
            log_coefficient_modulus=self.log_coefficient_modulus,
            modulus_degree=self.modulus_degree,
            message=row,
        )
    )

  def __str__(self) -> str:
    # this does not need to be fast because it will only be used in development.
    inner = '\n '.join(self._single_row_str(row) for row in self.message)
    return f'[\n {inner}\n]'

  def __repr__(self) -> str:
    return str(self)


@dataclasses.dataclass
class RlweSecretKey:
  """A secret key for the RLWE encryption scheme."""

  # the log of q in Z/qZ
  log_coefficient_modulus: int

  # the power of two N in the polynomial modulus x^N + 1
  modulus_degree: int

  # the number of samples chosen, equal to len(RlweCiphertext) - 1
  rlwe_dimension: int

  # the binary polynomial values (s_1, ..., s_{rlwe_dimension})
  # used as a dot product multiplicand when encrypting.
  # each entry is a polynomial in {0,1}[X]/ (X^N + 1)
  data: jnp.ndarray

  @property
  def coefficient_modulus(self) -> jnp.uint32:
    return jnp.uint32(2) ** self.log_coefficient_modulus


def gen_key(
    params: parameters.SchemeParameters, prg: random_source.RandomSource
) -> RlweSecretKey:
  """Generate an RLWE secret key."""
  return RlweSecretKey(
      log_coefficient_modulus=params.log_plaintext_modulus,
      modulus_degree=params.polynomial_modulus_degree,
      rlwe_dimension=params.rlwe_dimension,
      data=prg.sk_uniform(
          shape=(params.rlwe_dimension, params.polynomial_modulus_degree),
          dtype=jnp.uint32,
      ),
  )


def encrypt(
    plaintext: RlwePlaintext, sk: RlweSecretKey, prg: random_source.RandomSource
) -> RlweCiphertext:
  """Encrypt an RLWE plaintext."""
  ai_samples = prg.uniform(
      shape=(sk.rlwe_dimension, sk.modulus_degree),
      dtype=jnp.uint32,
  )
  error_sample = prg.rounded_normal(shape=(sk.modulus_degree,)).astype(
      jnp.uint32
  )
  ciphertext_data = jit_encrypt(
      plaintext.message,
      sk.data,
      ai_samples,
      error_sample,
      log_coefficient_modulus=sk.log_coefficient_modulus,
  )

  return RlweCiphertext(
      log_coefficient_modulus=sk.log_coefficient_modulus,
      modulus_degree=sk.modulus_degree,
      message=ciphertext_data,
  )


@functools.partial(jax.jit, static_argnames='log_coefficient_modulus')
def jit_encrypt(
    plaintext: jnp.ndarray,
    key_data: jnp.ndarray,
    ai_samples: jnp.ndarray,
    error_sample: jnp.ndarray,
    log_coefficient_modulus: int,
) -> jnp.ndarray:
  """Encrypt an RLWE plaintext with pre-computed randomness."""
  clean_product = (
      matrix_utils.poly_dot_product(
          ai_samples,
          key_data,
      )
      + plaintext
  )
  obfuscated_product = clean_product + error_sample

  if log_coefficient_modulus < 32:
    modulus = jnp.uint32(2) ** log_coefficient_modulus
    obfuscated_product = obfuscated_product % modulus

  return jnp.append(ai_samples, jnp.array([obfuscated_product]), axis=0)


def decrypt(
    ciphertext: RlweCiphertext,
    sk: RlweSecretKey,
    encoding_params: encoding.EncodingParameters,
) -> RlwePlaintext:
  """Decrypt an RLWE ciphertext."""
  dot_prod = matrix_utils.poly_dot_product(ciphertext.message[:-1], sk.data)
  obfuscated_plaintext = ciphertext.message[-1] - dot_prod
  if sk.log_coefficient_modulus < 32:
    obfuscated_plaintext = jnp.mod(obfuscated_plaintext, sk.coefficient_modulus)
  msg = encoding.remove_noise(obfuscated_plaintext, encoding_params)
  return RlwePlaintext(
      log_coefficient_modulus=sk.log_coefficient_modulus,
      modulus_degree=sk.modulus_degree,
      message=msg,
  )


@jax.named_call
def flatten_key(sk: RlweSecretKey) -> lwe.LweSecretKey:
  """Flattens rlwe secret key to use for decrypting after sample extraction.

  Args:
    sk: an RLWE secret key that was used to encrypt the input to a sample
      extract.

  Returns:
    An LWE secret key that can be used to decrypt the LWE ciphertext output of
    a sample extraction.
  """
  poly_deg, k = sk.modulus_degree, sk.rlwe_dimension
  return lwe.LweSecretKey(
      log_modulus=sk.log_coefficient_modulus,
      lwe_dimension=poly_deg * k,
      key_data=sk.data.reshape(poly_deg * k),
  )
